#include "gimps.h"
#include "nsis.h"
#include <stdio.h>

int hdr_nsis::offset_;

static class NSISHeaderClass : public PacketHeaderClass
{
public:
  NSISHeaderClass() : PacketHeaderClass("PacketHeader/NSIS",
                                            sizeof(hdr_nsis))
  {
    bind_offset (&hdr_nsis::offset_);
  }

}
class_NSIShdr;

static class GIMPSClass : public TclClass
{
public:
  GIMPSClass() : TclClass("Agent/GIMPS") {}
  TclObject* create(int, const char*const*)
  {
    return (new GIMPSAgent());
  }
}
class_GIMPS;

GIMPSAgent::GIMPSAgent() : Agent(PT_NSIS), GIMPS_timer_(this),
    ip6_(0)
{

  gimpsIpVer = IPV4; // Check with TlvObjectNA
  //lifetime = 30; //RSVT time (added to the NLI object)
  NSLPID = 1;
  bind("refresh_", &refresh_);
  bind("lifetime_", &lifetime);
  bind("ip6_", &gimpsIpVer);
  bind("nslpid_", &NSLPID);
  bind("poolingTime_", &poolingTime);
  bind("messageRoutingMethod_", &(messageRoutingMethod_));
  bind_bool("nam_", &nam_);
  bind("debug_", &debug);

  gimpsPortNumber = 999;
  gimpsSourcePrefix = 100001;
  gimpsVersion = 1;
  hopCount = 5;
  messageRoutingState = NULL;
  messageAssociationState = NULL;
  handlerNSLP = NULL;
  //poolingTime = 20; //the periodicity on which the MRSs will be checked
  gimpsInterfaces = 0;
  MTU = 9000;
  sessionIdcounter = 0;
  messageRoutingMethod_ = 1;
  listMessageGimps = NULL;
  peerIdentitysize = 128;
  transmissionMode = CMODE;
  fid_ = 0;
  receiver_tcp_app = NULL;
  sender_tcp_app = NULL;
  refreshment = 1;
  nslp_data_info = NULL;
  nslp_ = NULL;
  //debug = 0;
};

int GIMPSAgent::command(int argc, const char*const* argv)
{
  nsaddr_t n1;
  Tcl& tcl = Tcl::instance();
  
  if (strcmp(argv[1], "start") == 0)
  {
    GIMPS_timer_.resched(0.0);
    return (TCL_OK);
  };
  
  if (strcmp(argv[1], "stop-refreshment") == 0)
  {
  	setRefreshment(0);
    GIMPS_timer_.resched(0.0);
    return (TCL_OK);
  };
  
  if (strcmp(argv[1], "restart-refreshment") == 0)
  {
  	setRefreshment(1);
    GIMPS_timer_.resched(0.0);
    return (TCL_OK);
  };
  
  if (strcmp(argv[1], "setipgimps") == 0)
  {
    if (argc != 3)
    {
      printf("Error while you trying to setup Ip for gimps layer,\\ 
      please check the file ns-nsis.tcl setipgimps");
      return (TCL_ERROR);
    }
    tcl.evalf("%s id", argv[2]);   //here the ns2 provide a node object similar to _o13
    n1 = atoi(tcl.result());	   //so we need to invoke the ns to return us the node's ID.
    gimpsIP = n1;        		   //address of the node
    generatePeerIdentity(gimpsIP);
    setsession(gimpsIP);	//set the fist session
    return (TCL_OK);
  }

  if (!strcmp(argv[1], "refresh"))
  {
    refresh() ;
    return (TCL_OK);
  };
  if (!strcmp(argv[1], "config-mode"))
  {
    if(!strcmp(argv[2], "dmode"))
      transmissionMode = DMODE;
    else
      transmissionMode = CMODE;
    return (TCL_OK);
  };
  return (Agent::command(argc, argv));
};

char GIMPSAgent::give(Packet *p, NSISChecker *ret)
{
  unsigned int temp = 0;
  GIMPSMessage *msg = NULL;
  hdr_nsis* nsis_hdr = hdr_nsis::access(p);
  //hdr_cmn *cmn_hdr = hdr_cmn::access(p);
  memcpy(&temp, p->accessdata(), sizeof(p->accessdata()));
  msg = (GIMPSMessage *)temp;
  msg->fromhop = nsis_hdr->fromhop;
  
  this->receiveMessage(msg);
  if (msg->forward == 0)
  {
    //Packet::free(p);
    //printf("\n[GIMPSAgent - give]<Node %d> Packet destructed\n",this->gimpsIP);
    return 0;
  }
  prnt("give", "Packet going back to the link\n");
  return 1; 
};

void GIMPSAgent::reset_GIMPSAgent_Timer()
{
  GIMPS_timer_.resched((double)poolingTime);
}

void GIMPSAgent_Timer::expire(Event *e)
{
  agent_->refresh();
  agent_->reset_GIMPSAgent_Timer();
}

void GIMPSAgent::recv(Packet* p, Handler* h)
{
  unsigned int temp;
  hdr_cmn *cmn_hdr = hdr_cmn::access(p);
  hdr_ip *ip_hdr = hdr_ip::access(p);
  memcpy(&temp, p->accessdata(), sizeof(int));
  GIMPSMessage *msg = NULL;
  msg = (GIMPSMessage *)temp;
  msg->fromhop = cmn_hdr->last_hop_;
  
  if (ip_hdr->prio_ == -1) {
  	/*the packet was already processed by this node, since the nsis-Link routed
  	 *  it some steps ago. The packet was forced to arrived  
  	 * here by the conventional way just to be possible to track it without 
  	 * any problem by the NAM*/
  	 //prnt("recv","Deleting packet\n");
  	delete msg;
  	Packet::free(p);
  	return;
  }
  Packet::free(p);
  this->receiveMessage(msg);
  delete msg;
  /* All packets are already being processed by the 'give' function,
     so get rid of this packet. */
  // Packet::free(p);
};


void GIMPSAgent::recv(unsigned int ptr) 
{ //packet received through a TCP connection using TcpApp applications
	if (ptr == 0) {
		prnt("recv", "Error receiving message by TCP connection\n");
		exit(0);	
	}
	prnt("rcv", "> Message arrived by TCP connection\n");
	GIMPSMessage *msg = (GIMPSMessage *)ptr;
	this->receiveMessage(msg);
	delete msg;
};


void GIMPSAgent::refresh()
{
  //this->checkAllMRS(); 
}

void GIMPSAgent::setsession(nsaddr_t IP)
{
  sessionIdcounter = IP * 1000;
}

void GIMPSAgent::removemessage(GIMPSMessage *msg)
{
  if (msg == NULL)
  {
    prnt("removemessage", "Trying to remove a NULL message. Bye!\n");
    exit(0);
  }
  delete msg;
}

GimpsTcpApp *GIMPSAgent::createTcpConnection(nsaddr_t src, nsaddr_t dst) {
	GimpsTcpApp *sender_app, *receiver_app, *sat, *cur;
	sender_app = receiver_app = NULL;
	GIMPSAgent *peer_agent;
	Tcl& tcl = Tcl::instance();
	
    tcl.evalf("[Simulator instance] add-tcp-connection %d %d", src, dst);
    if (strlen(tcl.result()) == 0) {
      	prnt("createTcpConnection", "It wasn't possible to find a TCP agent\n");
      	return NULL;
    }
    sender_app = (GimpsTcpApp *)TclObject::lookup(tcl.result());
    if (sender_app == 0) {
      	prnt("createTcpConnection", "It wasn't possible to find a TCP agent\n");
      	return NULL;
    }
    receiver_app = sender_app->getPeer();
    peer_agent = receiver_app->getGimpsAgent();
    
    if (receiver_app == NULL) {
    	prnt("createTcpConnection", 
    	"It wasn't possible to get the peer (receiver) GimpsTcpApp\n");
      	return NULL;	
    }
    sender_app->next = NULL;
    if (sender_tcp_app == 0) {
      	sender_tcp_app = sender_app;   
    }
    else {
    	cur = sender_tcp_app;
    	while (cur != NULL){
      		sat = cur;
      		cur = cur->next; 	
    	}
    	sat->next = sender_app;
    }
    //receiver list
    if (peer_agent->receiver_tcp_app == NULL) {
      	peer_agent->receiver_tcp_app = receiver_app;   
    }
    else {
    	cur = peer_agent->receiver_tcp_app;
    	while (cur != NULL) {
      		sat = cur;
      		cur = cur->next; 	
    	}
    	sat->next = receiver_app;
    }
      
	return sender_app;
}

void GIMPSAgent::sendMessage(GIMPSMessage *msg)
{
  unsigned int temp;
  GimpsTcpApp *tcp_agent = NULL;
  GIMPSAgent *receiver_agent = NULL;
  GimpsTcpApp *receiver_app = NULL;
  size_ = msg->get_length();
  size_ = (int)size_/8;
  dst_.addr_ = msg->getDestinationAddress();
  fid_ = msg->gettypemsg();
  Packet *pkt = allocpkt();
  pkt->allocdata(sizeof(temp));
  temp = (unsigned int) msg;
  memcpy(pkt->accessdata(), &temp, sizeof(temp));
  hdr_ip *ip_hdr = hdr_ip::access(pkt); 
  ip_hdr->saddr() = here_.addr_; //SM
  ip_hdr->daddr() = msg->getDestinationAddress(); 
  
  if (msg->protocol  == CONNECTION) {
  	Packet::free(pkt); //just needed in UDP connections
  	MessageRoutingState *mrs = getRoutingState(msg->getTlvMRI(), 
  	                           msg->getNslpId(), msg->getTlvSI());
  	if (mrs == NULL) {
  		prnt("sendMessage", 
  		"It wasn't possible to find a MRS suitable to get a TCP connection\n");
      	exit(0);
  	}  
  	//tcp connections re-use
  	if ((tcp_agent = (GimpsTcpApp *)mrs->getAssociation()) != NULL) { 
  		prnt("sendMesage","Sending msg by TCP\n");
  		tcp_agent->sendmsg(size_, temp);
  		return;
  	}
  	prnt("sendMessage", "creating tcp connection\n");
  	tcp_agent = createTcpConnection(here_.addr_, msg->getDestinationAddress());
  	if (tcp_agent == NULL) {
  	  prnt("sendMessage", "It wasn't possible to find a TCP agent\n");
      exit(0);
  	}
  	if(debug) {
  		printf("[GIMPSAgent sendMessage] <Node %d> TCP association created (%d -> %d)\n",
    	gimpsIP, here_.addr_, msg->getDestinationAddress());
    	fflush(stdout);
  	}
  	//prnt("sendMessage", "A TCP agent was found! :-)");
  	//TODO: decRef() when the association is disposed by the MRS
  	tcp_agent->incRef();
  	mrs->setAssociation((TcpApp *)tcp_agent);
  	receiver_app = tcp_agent->getPeer();
  	receiver_agent = receiver_app->gimps_agent_;
  	receiver_agent->setMaReceiverApp(receiver_app, msg->getTlvMRI(), 
  	 msg->getNslpId(), msg->getTlvSI());
  	prnt("sendMesage","Sending msg by TCP\n");	 	
  	tcp_agent->sendmsg(size_, temp);
  }
  else {
 	send(pkt, 0);
  }
  if(debug)
  	printf("[GIMPSAgent sendMessage] Node %d - Packet sent at direction %d -> %d\n", 
  	ip_hdr->saddr(), ip_hdr->saddr(), ip_hdr->daddr());
}

void GIMPSAgent::setMaReceiverApp(GimpsTcpApp *tcp_app, TlvObjectMRI *mri, 
    int nslpid, TlvObjectSI *si) {
	mri->changeDirection();									
	MessageRoutingState *mrs = getRoutingState(mri, nslpid, si);
  	mri->changeDirection(); //recovering the original value
  	if(mrs == NULL) {
  		prnt("setMaReceiverApp", 
  		"It was not possible to find MA to set the association\n");	
  		return;
  	}
  	prnt("setReceiverApp", "Configuring MA with the Tcp association\n");
  	tcp_app->incRef();
  	mrs->setAssociation((TcpApp *)tcp_app);
}

int  GIMPSAgent::getnewsession()
{
  sessionIdcounter++;
  return sessionIdcounter;
};

void GIMPSAgent::sendNslpQueuedMessages() {
  GIMPSMessage *data_msg;
  NslpDataInfo *nslpi, *aux; 
  nslpi = this->nslp_data_info;
  
  while (nslpi != NULL) {
  	data_msg = createDataMessage(this->NSLPID, nslpi->getMRI(), 
  	           nslpi->getSI(), nslpi->getData());
  	aux = nslpi;
  	nslpi = nslpi->next;
  	delete aux;
  	if (data_msg != NULL) {
  	  prnt("sendNslpQueueMessages","sending NSLP data\n");
  	  this->sendMessage(data_msg);
  	}
  }
}

void GIMPSAgent::enqueueNslpMessage(NslpDataInfo *nslpinfo) {
	NslpDataInfo *sat, *prev;
	
	sat = prev = nslp_data_info;
	if(sat == NULL) {
		nslp_data_info = nslpinfo;
		return;	
	}
	sat = sat->next;
	while(sat){
		prev = sat;
		sat = sat->next;
	}
	prev->next = nslpinfo;
}
void GIMPSAgent::sendMessage(TlvObjectMRI *mri, char *data, TlvObjectSI *si)
{ //called by a NSLP application 
	GIMPSMessage *msg;
	if(getRoutingState(mri, this->NSLPID, si) == NULL) {
		sendQueryMessage(mri, data, si);
	}
	else {
		msg = createDataMessage(this->NSLPID, mri, si, data);
		if (msg != NULL) {
			this->sendMessage(msg);
  	  		prnt("sendNslpQueueMessages", "sending NSLP data\n");
		}
	}
}

void GIMPSAgent::sendQueryMessage(TlvObjectMRI *mri, char *data, TlvObjectSI *si)
{
  //the query message doesn't contain any NSLP data message
  GIMPSMessage *msg = createQueryMessage(mri, NULL, si);
  
  if (msg != NULL) this->sendMessage(msg);
  if (data != NULL) {
  	NslpDataInfo *nslpinfo = new NslpDataInfo(mri, data, si);
  	prnt("sendQueryMessage","enqueuing NSLP Data\n");
  	enqueueNslpMessage(nslpinfo);
  }
}

GIMPSMessage *GIMPSAgent::createQueryMessage(TlvObjectMRI *mri1, char *data, 
                          TlvObjectSI *si)
{
  TlvObjectSI *sii;
  MessageRoutingState *mrs;
  TlvObjectNLI *nli;
  TlvObjectMRI *mri;
  
  mri = new TlvObjectMRI(mri1->get_contents());
  if(this->gimpsIP == mri->getDestinationAddress()){
    delete mri;
    return NULL;
  }
  sii = new TlvObjectSI(si->get_contents());
  //creates the GIMPS header and append the MRI and SI objects
  GIMPSMessage *msg = new GIMPSMessage(this->NSLPID, mri, sii); 
  msg->header.type = TYPE_QUERY_MSG;
  if (data != NULL) {
  	TlvObjectNSLP *datax = new TlvObjectNSLP(data);
  	msg->addTlvNSLP(datax);
  }
  msg->configHopCount(hopCount);// Define HopCount
  
  switch(msg->getDirection())
  {
  case DOWNSTREAM:
    mrs = this->getRoutingState(mri, this->NSLPID, sii);
    
    msg->setdestinationaddr(mri->getDestinationAddress());
    msg->setdestinationport(mri->getDestinationPort());
    msg->setprotocol(DATAGRAM);
    
    if (mrs == NULL)
    {	
      if(debug)
      	printf("[GIMPSAgent -createQueryMessage] Node %d - No route found to node %d\n", 
      	this->gimpsIP, mri->getDestinationAddress());
      nli = this->createNLI(NULL);
      // Add Node Addressing so receiver can know this node
      msg->addTlv(nli);	   
    }
    else
    { 
       mrs->active = 1;
       msg->addTlv(createNLI(mrs));
    }
    msg->addTlv(this->createQueryCookie());
    if(isMessageAssociationRequired()) 
    { 
        msg->addTlv(this->createStackProposal());
        msg->addTlv(this->createSCD());
    }
      /*
      if (msg->header.length >= this->MTU)
      { //check if exist ma other old this message (new object to be created
        
        prnt("createQueryMessage", "lenght > this->MTU");
        msg->setdestinationaddr(mrs->downstreamPeerMASPointer->getPeerInterface());
        msg->setdestinationport(mrs->downstreamPeerMASPointer->getPort());
        msg->setprotocol(DATAGRAM);
      }
      else
      {
       
        if (mrs->active == 0) {
          prnt("createQueryMessage", "active ==0");
          msg->addTlv(createNLI(mrs)); 
          msg->addTlv(this->createQueryCookie());  		
          msg->addTlv(this->createStackProposal());			
        }
        }
        */
    break;
  case UPSTREAM:
    prnt("createQueryMessage", "Switch:UPSTREAM\n");
    if (msg->header.length >= this->MTU)
    {
      msg->setdestinationaddr(mrs->downstreamPeerMASPointer->getPeerInterface());
      msg->setdestinationport(mrs->downstreamPeerMASPointer->getPort());
      msg->setprotocol(CONNECTION);
    }
    else
    {
      msg->setdestinationaddr(mri->getDestinationAddress());	//otherwise use datagram
      msg->setdestinationport(mri->getDestinationPort());
      msg->setprotocol(DATAGRAM);
    }
    break;
  };
  return msg;
};

char GIMPSAgent::nslpIdExists(int nslpId)
{
  if (this->NSLPID == nslpId) return 1;
  return 0;
};

TlvObjectQC *GIMPSAgent::createQueryCookie()
{
  TlvObjectQC *tlvQC = new TlvObjectQC();
  return tlvQC;
};

TlvObjectRC *GIMPSAgent::createResponderCookie()
{
  TlvObjectRC *rc = new TlvObjectRC();
  return rc;
};

TlvObjectNLI *GIMPSAgent::createNLI(MessageRoutingState *mrs)
{
  TlvObjectNLI *NLI;
  int ipttl;
  if (mrs != NULL)
    ipttl = mrs->IPTTL;
  else
    ipttl = 0;
  NLI = new TlvObjectNLI(peerIdentitysize, ipttl, lifetime, 
                         gimpsIpVer, peerIdentity, gimpsIP);

  return NLI;
};

TlvObjectSCD *GIMPSAgent::createSCD()
{
  TlvObjectSCD *tlvSCD = new TlvObjectSCD(1, CONNECTION, 0);
  return tlvSCD;
};

TlvObjectSP *GIMPSAgent::createStackProposal()
{
  TlvObjectSP *tlvSP = new TlvObjectSP(1, "0101010101010101010101010101");
  return tlvSP;
};

void GIMPSAgent::generatePeerIdentity(nsaddr_t IP)
{
  peerIdentity = IP;
};

char GIMPSAgent::cmpMRS(MessageRoutingState *mrs1, MessageRoutingState *mrs2)
{
  if (mrs1 == NULL)
  	return 0;
  if ((mrs1->mri->getDestinationAddress() == mrs2->mri->getDestinationAddress()) &&
      (mrs1->mri->getSourceAddress()      == mrs2->mri->getSourceAddress())      &&
      (mrs1->sessionId->getSessionId()    == mrs2->sessionId->getSessionId())    &&
      (mrs1->mri->getDirection()    == mrs2->mri->getDirection()))
  {
    return 1;
  };
  return 0;
};

// this method can add and replace MRSs. If the object already exists, 
//than we are going to replace it for another one.

void GIMPSAgent::RoutingState_add(MessageRoutingState *mrs)
{
  MessageRoutingState *objmrs1, *prev;
  objmrs1 = prev = this->messageRoutingState;
  //int i = 0;
  
  mrs->next = NULL;
  if (this->messageRoutingState == NULL)
  {
    this->messageRoutingState = mrs;
    return;
  }
  if (cmpMRS(this->messageRoutingState, mrs))  
  {     //the first of the list
        this->messageRoutingState = mrs;
        mrs->next = objmrs1->next;
        return;
  }
  objmrs1 = objmrs1->next;
  while (objmrs1 != NULL)
  {
    if (cmpMRS(objmrs1, mrs))
    {
      prev->next = mrs;
      mrs->next = objmrs1->next;
      delete objmrs1;
      return;
    }
    prev = objmrs1;
    objmrs1 = objmrs1->next;
  }
  prev->next = mrs;
};

MessageAssociationState *GIMPSAgent::lookUpAssociation(TlvObjectNLI *nli, 
char direction) {
  MessageRoutingState *mrs = this->messageRoutingState;
  MessageAssociationState *ma = NULL;
  /*it just checks the address and peeridentity of the NLI object*/
   
  while (mrs != NULL)
  {
  	if (direction == UPSTREAM && mrs->mri->getDirection() == UPSTREAM)
  		ma = mrs->getUpstreamPeerMA();
  	if (direction == DOWNSTREAM && mrs->mri->getDirection() == DOWNSTREAM)
  		ma = mrs->getDownstreamPeerMA();
  	if (ma != NULL) {	
    	if ( (ma->getPeerIdentity() == nli->getPeerIdentity()) &&
         (ma->getPeerInterface() == nli->getInterfaceAddress()) )
         //TODO: compare nslpip?
    	{
      		return ma;
    	}
  	}
    mrs = mrs->next;
  }
  return NULL;
}

MessageRoutingState *GIMPSAgent::getRoutingState(TlvObjectMRI *mri, int nslpId, TlvObjectSI *si)
{
  MessageRoutingState *objmrs = this->messageRoutingState;

  while (objmrs != NULL)
  {
    if ( (objmrs->mri->getDestinationAddress() == mri->getDestinationAddress()) &&
         (objmrs->mri->getSourceAddress() == mri->getSourceAddress()) &&
         (objmrs->sessionId->getSessionId() == si->getSessionId()) &&
         (objmrs->mri->getDirection() == mri->getDirection()) )
         //TODO: compare nslpip, Flow direction 
    {
      return objmrs;
    };
    objmrs = objmrs->next;
  }
  return NULL;
};

MessageRoutingState *GIMPSAgent::getRoutingState2(TlvObjectMRI *mri, int nslpId, TlvObjectSI *si)
{
  MessageRoutingState *objmrs = this->messageRoutingState;
  MessageRoutingState *tail = this->messageRoutingState;
  
  while (objmrs != NULL)
  {
    if ( (objmrs->mri->getDestinationAddress() == mri->getDestinationAddress()) &&
         (objmrs->mri->getSourceAddress() == mri->getSourceAddress()) &&
         (objmrs->sessionId->getSessionId() == si->getSessionId()) &&
         (objmrs->mri->getDirection() == mri->getDirection()) )
         //TODO: compare nslpip, Flow direction 
    {
      prnt("getRoutingState2", "MRS entry found\n");
      return objmrs;
    }
    tail = objmrs;
    objmrs = objmrs->next;
  }
  return NULL;
};

void GIMPSAgent::RoutingState_del(MessageRoutingState *mrs)
{
  MessageRoutingState *objmrs, *tail;
  objmrs = tail = this->messageRoutingState;
  GimpsTcpApp *tcp_app;
  
  if(cmpMRS(this->messageRoutingState, mrs)) 
  {
  	this->messageRoutingState = this->messageRoutingState->next;
  	delete mrs;
  	mrs = NULL; 
  	return;
  } 
  while (objmrs != NULL)
  {
    if (cmpMRS(objmrs, mrs))
    {
      tail->next = mrs->next;
      tcp_app = (GimpsTcpApp *)objmrs->getAssociation();
      tcp_app->decRef();
      objmrs->delAssociation();
      delete objmrs;
      mrs = NULL;
      return;
    }
    tail = objmrs;
    objmrs = objmrs->next;
  }
};

void GIMPSAgent::checkAllMRS()
{
  char del = 0;
  MessageRoutingState *tmp;
  MessageRoutingState *mrs = this->messageRoutingState;
  prnt("checkAllMRS", "checking refresh...");
  while (mrs != NULL)
  {
    if (mrs->hasDownstreamPeerInfo()) //if we dont have downstream, 
    {                            //so we dont need to send mensagem to refresh	
      if(debug)
      	printf("[GIMPSAgent - checkAllMRS] Node %d - Downstream timer: %d\n", 
      	this->gimpsIP, mrs->getDownstreamTimer());								
      if (mrs->getDownstreamTimer() <= 10) 
      { // Check if it has reached the max retry number				
        if (mrs->getDownstreamRetry() < mrs->maxRetry)	
        { //Can still resend the retry message
        	if(getRefreshment() == 0) {
        		prnt("checkAllMRS", "REFRESH message will not be sent\n");
        	}
        	else {
        		prnt("checkAllMRS", "preparing REFRESH message\n");
        		GIMPSMessage *msg = createQueryMessage(mrs->getMri(), NULL, 
        						mrs->getSessionId());	
            	if(isMessageAssociationRequired()) {
            	//if(mrs->getDownstreamPeerMA() != NULL) {
          			msg->setdestinationaddr(
          	 		mrs->getDownstreamPeerMA()->getPeerInterface());  
          		}
          		else
          			msg->setdestinationaddr(mrs->getDownstreamPeer());
          		this->sendMessage(msg);
        	}
          	mrs->addDownstreamRetry();
        } 
        else
        { // Max retry reached. Delete Downstream state
          nsaddr_t peer;
          if(!isMessageAssociationRequired())
          	peer = mrs->getDownstreamPeer();
          else
            peer = (mrs->getDownstreamPeerMA())->getPeerInterface();
          if(debug)
          	printf("[GIMPSAgent] Node %d - Deleting downstream MRS for node %d\n ", 
          	this->gimpsIP, peer);
          mrs->delAssociation();
          mrs->deleteDownstreamInfo(); //delete MAS	
          
          if (!mrs->hasUpstreamPeerInfo())
          {
            tmp = mrs; del = 1; // Check if MRS have Upstream msg
            mrs = mrs->next;
            // Remove this MRS State (number of retry exceeded)
            this->RoutingState_del(tmp); 
          }
        }
      }
      else
      { 	//Timer still valid
        mrs->decreaseDownstreamTimer(poolingTime);    
      }
    }
    else
    { //upstream MRS info
      if(debug)
      	printf("[GIMPSAgent checkAllMRS] Node %d - upstreamTimer %d\n", 
      	this->gimpsIP, mrs->getUpstreamTimer());
      if (mrs->hasUpstreamPeerInfo())
      {  							//ONLY HAVE UPSTREAM PEER (no refresh)
        if (mrs->getUpstreamTimer() < 1)
        {     						
          //Timer has ended - Delete upstream info and 
          //association (tcp connection)  
          nsaddr_t peer;
          if(!isMessageAssociationRequired())
          	peer = mrs->getUpstreamPeer();
          else
            peer = (mrs->getUpstreamPeerMA())->getPeerInterface();
          if(debug)
          	printf("[GIMPSAgent checkAllMRS] Node %d - Removing upstream MRS for node %d\n", 
          	this->gimpsIP, peer);
          mrs->delAssociation();
          mrs->deleteUpstreamInfo();		
          if (!mrs->hasDownstreamPeerInfo())
          {
            tmp = mrs; del = 1;			//Check if MRS have Downstream msg
            mrs = mrs->next;
            this->RoutingState_del(tmp);
            prnt("checkAllMRS", "Upstream MRS removed\n");
          }
        }
        else
          mrs->decreaseUpstreamTimer(poolingTime);
      }
      else
      {
      	//never reached?
        tmp = mrs; del = 1;
        mrs = mrs->next;
        if(debug)
        	printf("[GIMPSAgent checkAllMRS] Node %d - (never) upstream MRS removed for node %d\n", 
            	this->gimpsIP, tmp->getDownstreamPeer());
        this->RoutingState_del(tmp); 	// Remove this Message Routing State
      }
    }
    if(!del)
      mrs = mrs->next;
    del = 0;  
  }
  //printf("[GIMPSAgent - checkAllMRS] END method\n");
};

void GIMPSAgent::receiveMessage(GIMPSMessage *msg)
{
  msg->forward = 0;             
  if (msg->receivebydatagram()) 			// Sent throught datagram
  {
  	//shouldn't this condition come before the above one? checkit
    if (!this->nslpIdExists(msg->getNslpId()))	
    {
      msg->decreaseHopCounter();   			// Decrease HopCount
      this->forwardmessage(msg);			// FORWARD MESSAGE
      return;
    }
    processMessage(msg);
  }
  else
  {
    // Verify the Routing State table to ensure this messaging association 
    //is associated with the MRI/NSLPID combination
    TlvObjectMRI *mri = new TlvObjectMRI(msg->getTlvMRI()->get_contents());
    mri->changeDirection();
    if (getRoutingState(mri, msg->getNslpId(), msg->getTlvSI()) != NULL){
      delete mri;
      processMessage(msg);   			// Send message for processing
    }
    else {
      printf("[GIMPS receiveMessage] Node %d - No Rounting State available (Connection), discarding message",
      gimpsIP);
      delete mri;
    }  
  }
};

unsigned char GIMPSAgent::isMessageAssociationRequired()
{
	//TODO: the decision if a MA is required or not depends on:
	//1) the msg protocol type:DATAGRAM or CONNECTION
	//2) The MRI attributes (?)
	//3) the policies of the nodes, definded by the user on tcl level 
	//if(msg->getTlvMRI()->getProtocol() == CONNECTION)
	if(transmissionMode == CMODE)
	  return 1;
	return 0;
}

int GIMPSAgent::compareCookies(GIMPSMessage *msg){
	TlvObjectQC *cookie;
	cookie = msg->getTlvQC();
	if (!cookie)
		return 0;
	//TODO:to compare with the current cookie
	delete cookie;	
	return 1;	
}

void GIMPSAgent::processMessage(GIMPSMessage *msg)
{
  char discardMsg = 0;
  char direction = 0;
  MessageRoutingState *mrs_;
  TlvObjectMRI *mri;
  //TlvObjectSI *si;
  MessageAssociationState *ma_;
  TlvObjectNLI *nli;
  
  prnt("processMessage", ">> Message arrived\n");
  /*
  if ( msg->getDirection() != DOWNSTREAM) 
  { 	// Query or Responder cookie?
  	//TODO
	return;
  }
  */
  switch (msg->gettypemsg())
  {
  case TYPE_QUERY_MSG:
    //This means:
    //1) refresh message 
    //2) requirement to stablish a D'Mode connection
    //3) requirement to stablish a C'Mode connection
    //TYPE_DATA_MSG: a Query encapsulated Data message 
    //(in case where no routing states are needed to send it)
    prnt("processMessage", "Message type = TYPE_QUERY_MSG\n");
  	mri = new TlvObjectMRI(msg->getTlvMRI()->get_contents());
    mri->changeDirection(); //The RS table maintain entries in reversed directions
    direction = mri->getDirection();
    mrs_ = getRoutingState(mri, msg->getNslpId(), msg->getTlvSI());
    delete mri;
    if(mrs_ != NULL)
    {   //a refresh message
		prnt("processMessage", "> Refresh message arrived\n");
		refresh_mrs_time(msg);
		if(mrs_->upstreamPeerMASPointer == NULL) //TODO: a new method to search the MAS!
		{ //D'Mode
	  	  mrs_->resetUpstreamTimer();
	  	  GIMPSMessage *gm = createResponseMessage(msg, mrs_, DATAGRAM);
		  this->sendMessage(gm);
		}
		else{
			GIMPSMessage *gm = createResponseMessage(msg, mrs_, CONNECTION);
			this->sendMessage(gm);
    	}
    }
    else { //A new connection
      if(!isMessageAssociationRequired())
      {   //To stablish a D'Mode connection 
		mrs_ = new MessageRoutingState(msg, -1, 0); //MA is NULL
        this->RoutingState_add(mrs_);
		///GIMPSMessage *gmesg = createResponseMessage(msg, mrs_, DATAGRAM);
		if(debug)
			printf("[GIMPSAgent processMessage] Node %d - New D'Mode Connection to node %d\n", 
			this->gimpsIP, msg->getTlvNLI()->getInterfaceAddress());
      }
      else 
      {//to Stablish a C'Mode connection
    	nli =  msg->getTlvNLI();
    	//Association Re-use
    	ma_ = lookUpAssociation(nli, direction);
    	if (ma_ == NULL) {
    		ma_ = new MessageAssociationState(nli, nli->getRSVT());
    		prnt("processMessage", "no TCP connection to be reused\n");
    	}
    	else
    		prnt("processMessage", "Reusing an already established TCP connection\n");
    	mrs_ = new MessageRoutingState(msg, ma_, -1, 0);
    	this->RoutingState_add(mrs_);
    	if(debug)
    		printf("[GIMPSAgent processMessage] Node %d - New C'Mode Connection to node %d\n",
    		this->gimpsIP, msg->getTlvNLI()->getInterfaceAddress());
      }
      GIMPSMessage *gmesg = createResponseMessage(msg, mrs_, DATAGRAM);		
      this->sendMessage(gmesg); //upstream message
      //downstream message
      sendQueryMessage(msg->getTlvMRI(), NULL, msg->getTlvSI());
    }
    break;
  case TYPE_RESPONSE_MSG: //upstream
      prnt("processMessage", "Message Type = TYPE_RESPONSE_MSG\n");
      ///getchar();
      if(!this->compareCookies(msg)){
      	prnt("processMessage", "QUERY COOKIE is unknown\n");
      	msg->forward = 0;
      	//TODO: to send error message
      	break;
      }
      nli = msg->getTlvNLI();
      mri = new TlvObjectMRI(msg->getTlvMRI()->get_contents());
      mri->changeDirection();
	  // Set upstream peer or Refresh upstream timer 
      mrs_ = getRoutingState(mri, msg->getNslpId(), msg->getTlvSI());
      direction = mri->getDirection();
      delete mri;
      if (mrs_)
      {
      	//refresh message
      	mrs_->resetDownstreamRetry();
      	mrs_->resetDownstreamTimer();
      	prnt("processMessage", "REFRESH msg arrived. Retry Timer reseted\n");
      }
      else
      { 
      	//it's a response to a Query asking for establishment of a connection
      	if(!isMessageAssociationRequired())
      	{
      		mrs_ = new MessageRoutingState(msg, -1, 1);
        	this->RoutingState_add(mrs_);
        	if(debug)
        		printf("[GIMPSAgent processMessage] Node %d - Sender: New D'Mode Connection to node %d\n", 
        		this->gimpsIP, msg->getTlvNLI()->getInterfaceAddress());
        	GIMPSMessage *gmesg = createConfirmMessage(msg, mrs_, DATAGRAM);
            this->sendMessage(gmesg);
        	sendNslpQueuedMessages();
      	}
      	else 
      	{
      		ma_ = lookUpAssociation(nli, direction);
    		if (ma_ == NULL) {
    			ma_ = new MessageAssociationState(nli, nli->getRSVT());
    			prnt("processMessage", "no TCP connection to be reused\n");
    		}
    		else
    			prnt("processMessage", "Reusing an already established TCP connection\n");
      		mrs_ = new MessageRoutingState(msg, ma_, -1, 1);
        	this->RoutingState_add(mrs_);
      		long ii = msg->getRSVT();
      		mrs_->setDownstreamPeer(ma_, ii);
      		if(debug)
      			printf("[GIMPSAgent processMessage] Node %d - Sender: New C'Mode Connection to node %d\n", 
        		this->gimpsIP, nli->getInterfaceAddress());
      		GIMPSMessage *gmesg = createConfirmMessage(msg, mrs_, CONNECTION);
            this->sendMessage(gmesg);
            //verify the NSLP Data messages queue
            sendNslpQueuedMessages();
      	}
      }
      break;
    case TYPE_DATA_MSG:
   		prnt("processMessage", "Message Type = TYPE_DATA_MSG\n");
   		//send data to nslp layer
   		refresh_mrs_time(msg);
    	if (msg->hasNslp() && !discardMsg) { addNSLPmessage(msg); }  
  		break;	  
    case TYPE_CONFIRM_MSG:
      prnt("processMessage", "Message Type = TYPE_CONFIRM_MSG\n");
      ///getchar();
      TlvObjectMRI *mri = new TlvObjectMRI(msg->getTlvMRI()->get_contents());
      mri->changeDirection();
      mrs_ = getRoutingState(mri, msg->getNslpId(), msg->getTlvSI());//Check MRS for this message.
      delete mri;
      if (mrs_ == NULL)
      {
        printf("GIMPSProcessor: No RoutingState for the received message (no QC nor RC) - DISCARDING MSG");
        discardMsg = 1;
      }
      else { 
      	refresh_mrs_time(msg);
      }
      break;
  } //end switch 
  if(discardMsg) {
  	//TODO: discard message
  }
}

void GIMPSAgent::forwardmessage(GIMPSMessage *msg)
{
  //that is.. we dont need to do nothing because the nsis agent is not present
  //at the current node
  msg->forward = 1;					
}								

void GIMPSAgent::refresh_mrs_time(GIMPSMessage *msg)
{
  TlvObjectMRI *mri = new TlvObjectMRI(msg->getTlvMRI()->get_contents());
  //TlvObjectSI *si;
  mri->changeDirection();
  MessageRoutingState *mrs_ = getRoutingState(mri, msg->getNslpId(), msg->getTlvSI());
  if(mri->getDirection() == UPSTREAM) {
  	//mrs_->setupstreamtime(msg->getRSVT());
  	mrs_->resetUpstreamTimer();
  	prnt("refresh_mrs_time", "Upstream Timer refreshed\n");
  }
  if(mri->getDirection() == DOWNSTREAM) {
  	//mrs_->setdownstreamtime(msg->getRSVT());
  	mrs_->resetDownstreamTimer();
  	prnt("refresh_mrs_time", "Downstream Timer refreshed\n");
  }
  delete mri;
}

GIMPSMessage *GIMPSAgent::createResponseMessage(GIMPSMessage *oldmsg, 
              MessageRoutingState *mrs_, int conn_type)
{
  TlvObjectNSLP *nlspData = NULL;
  TlvObjectMRI *mri = new TlvObjectMRI(oldmsg->getTlvMRI()->get_contents());
  mri->changeDirection();
  TlvObjectSI *sessionId = new TlvObjectSI(oldmsg->getTlvSI()->get_contents());
  int nslpId = oldmsg->getNslpId();
  TlvObjectQC *qc = oldmsg->getTlvQC();
  //to echo MRI, SI, and Query-Cookie to the querying node
  //A Response messages can include a NSLP payload depending on 
  //the NSLP application
  GIMPSMessage *msg = new GIMPSMessage(nslpId, mri, sessionId, nlspData);
  
  msg->addTlv(qc);
  msg->settypemsg(TYPE_RESPONSE_MSG);
  TlvObjectNLI *nli = oldmsg->getTlvNLI();
  
  msg->setdestinationaddr(nli->getInterfaceAddress());
  msg->configHopCount(oldmsg->getgimpshops());
  TlvObjectNLI *nli2 = createNLI(mrs_);
  msg->addTlv(nli2);
  msg->setprotocol(conn_type);
  msg->addTlv(createResponderCookie());
  if(!isMessageAssociationRequired()) //it's a D'Mode connection
    return msg;
  //NLI should go only on a D'Mode connection i.e., before 
  //the connection is set 
  msg->addTlv(createStackProposal());
  msg->addTlv(createSCD());
  return msg;
}

GIMPSMessage *GIMPSAgent::createDataMessage(int nslpId, TlvObjectMRI *mri, 
			  TlvObjectSI *sessionId, char *nslpData)
{
  MessageRoutingState *mrs;
  TlvObjectMRI *mri_;
  TlvObjectSI *sid;
  nsaddr_t dst;
  TlvObjectNSLP *data;
  
  mri_ = new TlvObjectMRI(mri->get_contents());
  sid = new TlvObjectSI(sessionId->get_contents());
  data = new TlvObjectNSLP(nslpData);
  GIMPSMessage *msg = new GIMPSMessage(nslpId, mri_, sid, data);
  msg->settypemsg(TYPE_DATA_MSG);
  mrs = getRoutingState(mri, nslpId, sessionId);
  if (mrs == NULL)
  {
  	prnt("createDataMessage", 
  	"it was not possible to find a route to transmit the NSLP data message");
  	return NULL;
  }
  if(isMessageAssociationRequired()) {
  	if (mrs->getDownstreamPeerMA() == NULL) {
  		prnt("createDataMessage", 
  		"it was not possible to find a MAS to transmit the NSLP data message");
  		return NULL;
  	}
  	dst = (mrs->getDownstreamPeerMA())->getPeerInterface();
  }
  else
  	dst = mrs->getDownstreamPeer();
  msg->setdestinationaddr(dst);
  msg->configHopCount(hopCount);
  if(isMessageAssociationRequired())
  	msg->setprotocol(CONNECTION);
  else
  	msg->setprotocol(DATAGRAM); 
  return msg;
}

void GIMPSAgent::sendQueryResponse(GIMPSMessage *oldmsg, MessageRoutingState *mrs_)
{
  //to echo MRI, SI, and Query-Cookie to the querying node
  TlvObjectNSLP *nlspData = NULL;
  TlvObjectMRI *mri = new TlvObjectMRI(oldmsg->getTlvMRI()->get_contents());
  mri->changeDirection();
  // MRI, SessionID
  TlvObjectSI *sessionId = new TlvObjectSI(oldmsg->getTlvSI()->get_contents());  
  int nslpId =oldmsg->getNslpId();
  TlvObjectQC *qc = oldmsg->getTlvQC();
  GIMPSMessage *msg = new GIMPSMessage(nslpId, mri, sessionId, nlspData);
  //the NLI of the receiving should be sent instead! checkit
  TlvObjectNLI *nli = new TlvObjectNLI(oldmsg->getTlvNLI()->get_contents());
  msg->setdestinationaddr(nli->getInterfaceAddress());
  msg->setprotocol(DATAGRAM);
  msg->configHopCount(oldmsg->getgimpshops());// Define HopCount
  msg->addTlv(qc);				// Process the query and create RESPONDER COOKIE
  msg->addTlv(createResponderCookie());	// Process the query and create RESPONDER COOKIE
  msg->addTlv(createNLI(mrs_)); //Add Node Addressing so receiver can know this node
  msg->addTlvMRI(mri);				// Add MRI to msg
  msg->addTlv(createStackProposal());
  msg->addTlv(createSCD());
  this->sendMessage(msg); 
}

void GIMPSAgent::sendQueryCookiebyMA(GIMPSMessage *oldmsg, MessageRoutingState *mrs_)
{
  MessageAssociationState *ma_ = mrs_->upstreamPeerMASPointer;
  TlvObjectNSLP *nlspData = NULL;
  TlvObjectMRI *mri = new TlvObjectMRI(oldmsg->getTlvMRI()->get_contents()); // MRI, SessionID
  TlvObjectSI *sessionId = new TlvObjectSI(oldmsg->getTlvSI()->get_contents());
  int nslpId =oldmsg->getNslpId();
  TlvObjectQC *qc = oldmsg->getTlvQC();
  mri->changeDirection();
  GIMPSMessage *msg = new GIMPSMessage(nslpId,   mri, sessionId, nlspData);
  msg->setdestinationaddr(ma_->getPeerInterface());
  msg->setdestinationport(ma_->getPort());
  msg->setprotocol(CONNECTION);
  msg->configHopCount(hopCount);					// Define HopCount
  msg->addTlv(qc);	
  msg->addTlv(createNLI(mrs_)); // Add Node Addressing so receiver can know this node
  msg->addTlvMRI(mri);						// Add MRI to msg
  this->sendMessage(msg); 					// send object
}

GIMPSMessage *GIMPSAgent::createConfirmMessage(GIMPSMessage *oldmsg, 
              MessageRoutingState *mrs_, int conn_type)
{
  MessageAssociationState *ma_;
  TlvObjectRC *rc;
  TlvObjectNSLP *nlspData = NULL;
  TlvObjectMRI *mri = new TlvObjectMRI(oldmsg->getTlvMRI()->get_contents()); 
  TlvObjectSI *sessionId = new TlvObjectSI(oldmsg->getTlvSI()->get_contents());
  mri->changeDirection();
  int nslpId = oldmsg->getNslpId();
  
  GIMPSMessage *msg = new GIMPSMessage(nslpId, mri, sessionId, nlspData);
  if(isMessageAssociationRequired()) {
  	ma_ = mrs_->downstreamPeerMASPointer;
  	msg->setdestinationaddr(ma_->getPeerInterface());
  	msg->setdestinationport(ma_->getPort());
  }
  else {
  	msg->setdestinationaddr(mrs_->getDownstreamPeer());
  	//msg->setdestinationport(mrs_->getPort());
  }
  msg->setprotocol(conn_type);
  msg->settypemsg(TYPE_CONFIRM_MSG);
  msg->configHopCount(hopCount);					// Define HopCount
  rc = oldmsg->getTlvRC();  
  msg->addTlv(rc);
  msg->addTlv(createNLI(mrs_)); // Add Node Addressing so receiver can know this node
  return msg;
}

void GIMPSAgent::sendQueryConfirm(GIMPSMessage *oldmsg, MessageRoutingState *mrs_)
{
  MessageAssociationState *ma_ = mrs_->downstreamPeerMASPointer;
  TlvObjectNSLP *nlspData = NULL;
  TlvObjectMRI *mri = new TlvObjectMRI(oldmsg->getTlvMRI()->get_contents());
  TlvObjectSI *sessionId = new TlvObjectSI(oldmsg->getTlvSI()->get_contents());
  mri->changeDirection();
  int nslpId = oldmsg->getNslpId();
  TlvObjectRC *rc = oldmsg->getTlvRC();
  GIMPSMessage *msg = new GIMPSMessage(nslpId, mri, sessionId, nlspData);
  msg->setdestinationaddr(ma_->getPeerInterface());
  msg->setdestinationport(ma_->getPort());
  msg->setprotocol(CONNECTION);
  msg->settypemsg(TYPE_CONFIRM_MSG);
  msg->configHopCount(hopCount);					// Define HopCount  
  msg->addTlv(rc);
  msg->addTlv(createNLI(mrs_)); 
  this->sendMessage(msg); 
}

void GIMPSAgent::MessageAssociation_add(MessageAssociationState *ma)
{
  MessageAssociationState *objMA = this->messageAssociationState;
  MessageAssociationState *tail  = this->messageAssociationState;
  while (objMA != NULL)
  {
    tail=objMA;
    objMA=objMA->next;
  }
  tail->next=ma;
  ma->prev=tail;
  ma->next = NULL;
}

MessageAssociationState *GIMPSAgent::getMessageAssociation(int p_Identity)
{
  MessageAssociationState *objMA = this->messageAssociationState;
  MessageAssociationState *tail  = this->messageAssociationState;

  while (objMA != NULL)
  {
    if (objMA->getPeerIdentity() == p_Identity)
      return objMA;
    tail=objMA;
    objMA=objMA->next;
  }
  return NULL;
}

void GIMPSAgent::MessageAssociation_del(int p_Identity)
{

  MessageAssociationState *objMA = this->messageAssociationState;
  MessageAssociationState *tail  = this->messageAssociationState;
  //MessageAssociationState *ma;
  while (objMA != NULL)
  {
    if (objMA->getPeerIdentity()==p_Identity)
    {
      tail->next= objMA->next;
      objMA->next->prev =tail;
      delete objMA;
      return ;
    }
    tail=objMA;
    objMA=objMA->next;

  }
}

void GIMPSAgent::addNSLPmessage(GIMPSMessage *msg)
{
  TlvObjectMRI *mri        = new TlvObjectMRI(msg->getTlvMRI()->get_contents());
  TlvObjectSI *si          = new TlvObjectSI(msg->getTlvSI()->get_contents());
  TlvObjectNSLP *NSLPData  = msg->getTlvNSLP();
  char directionflow       = msg->getDirection();
  int NslpID               = msg->getNslpId();

  NSLPmessage *nq = new NSLPmessage(mri , si, NSLPData, directionflow, 
                                    NslpID, msg->fromhop);
  prnt("addNSLPmessage", "Delivering message to NSLP Layer\n");
  nslp_->recvnslp(nq);						//send message to nslp layer
  //this->removemessage(msg);
};

NSLPmessage *GIMPSAgent::getNSLPmessage()
{
  NSLPmessage *tmp=handlerNSLP;
  if (handlerNSLP !=NULL)
    handlerNSLP= handlerNSLP->next;
  return tmp;
};

void GIMPSAgent::add_message_for_agent(GIMPSMessage *ma)
{
  GIMPSMessage *objMA = this->listMessageGimps;
  GIMPSMessage *tail  = this->listMessageGimps;
  while (objMA != NULL)
  {
    tail=objMA;
    objMA=objMA->next;
  }
  if (this->listMessageGimps==NULL)
  {
    this->listMessageGimps = ma;
    ma->prev= NULL;
    ma->next =NULL;
  }
  else
  {
    tail->next=ma;
    ma->prev=tail;
    ma->next = NULL;
  }
};

NSLPmessage::NSLPmessage(TlvObjectMRI *mri_ , TlvObjectSI *si_, 
             TlvObjectNSLP *NSLPData_, char directionflow_, 
             int NslpID_, nsaddr_t fromnode)
{
  mri = mri_;
  si = si_;
  NSLPData = NSLPData_;
  directionflow = directionflow_;
  NslpID = NslpID_;
  fromhop = fromnode;
};

void GIMPSAgent::prnt(char *method, char *msg)
{
	if (debug == 1) {
		printf("[GimpsAgent %s] Node %d - %s", method, this->gimpsIP, msg);
		fflush(stdout);
	}
}

NslpDataInfo::NslpDataInfo(TlvObjectMRI *mri, char *data, TlvObjectSI *si) {
	this->mri = new TlvObjectMRI(mri->get_contents());
	this->si = new TlvObjectSI(si->get_contents());
	this->data = new char [strlen(data)+1];
  	strcpy(this->data, data);
  	next = NULL;
}

NslpDataInfo::~NslpDataInfo() {
	delete this->mri;
	delete this->si;
	delete[] this->data;
}
