#ifndef gimps_h
#define gimps_h
#include "packet.h"
#include "ip.h"
#include "agent.h"
#include "address.h"
#include "adc.h"
//#include "gimps-tcp-app.h"
#include "gimpsmessage.h"
#include "nsis-link.h"

#include "webcache/tcpapp.h"
#include "app.h"

class GimpsTcpApp;


class NSLPApp;
class   TcpSinkNSIS;


struct hdr_nsis
{
  static int offset_ ;    // offset for this header
  inline static int& offset() { return offset_ ;}
  inline static hdr_nsis *access (const Packet *p)
  {
    return (hdr_nsis*) p->access(offset_) ;
  }

  nsaddr_t fromhop;
};

class GIMPSAgent;							//forward declarion
class NSLPmessage;
class NslpDataInfo;

class GIMPSAgent_Timer: public TimerHandler
{
public:
  GIMPSAgent_Timer(GIMPSAgent *agent): TimerHandler()
  {
    agent_ = agent;
  };
protected:
  GIMPSAgent *agent_;
  virtual void expire(Event *e);
};

class GIMPSAgent : public Agent
{
  //timer for sending package and send refresh message	
  friend class GIMPSAgent_Timer;
  public:
  GIMPSAgent();
  int command(int argc, const char*const* argv);
  void recv(Packet* p, Handler* h);
  void recv(unsigned int p);
  char give(Packet *p, NSISChecker *ret);
  void refresh();
   ~GIMPSAgent(){ };
  void setsession(nsaddr_t IP);
  void checkAllMRS();
  char nslpIdExists(int nslpId);
  void generatePeerIdentity(nsaddr_t IP);
  void receiveMessage(GIMPSMessage *msg);
  NSLPmessage *getNSLPmessage();
  GIMPSMessage *createQueryMessage(TlvObjectMRI *mri, char *data,  TlvObjectSI *si);
  void sendQueryMessage(TlvObjectMRI *mri, char *data, TlvObjectSI *si);
  int  getnewsession();
  void sendMessage(GIMPSMessage *msg);
  void sendMessage(TlvObjectMRI *mri, char *data, TlvObjectSI *si);
  void removemessage(GIMPSMessage *msg);
  unsigned char GIMPSAgent::isMessageAssociationRequired();
  int compareCookies(GIMPSMessage *msg);
  void GIMPSAgent::setMaReceiverApp(GimpsTcpApp *tcp_app, TlvObjectMRI *mri, 
  int nslpid, TlvObjectSI *si);
  MessageAssociationState *lookUpAssociation(TlvObjectNLI *nli, char direction);
  void setRefreshment(int status) { refreshment = status; }
  int getRefreshment() {return refreshment; }
  // General Variables
  nsaddr_t gimpsIP;
  int gimpsPortNumber ;
  int gimpsIpVer;  						// Check with TlvObjectNA
  int gimpsSourcePrefix ;
  int gimpsVersion ;
  int hopCount;
  MessageRoutingState     *messageRoutingState;
  MessageAssociationState *messageAssociationState;
  MessageRoutingState  *mmm;
  NSLPmessage *handlerNSLP;
  int gimpsInterfaces;
  int lifetime;
  int MTU;
  int NSLPID;
  GIMPSMessage *listMessageGimps; 
  int peerIdentity;
  int  peerIdentitysize;
  int  messageRoutingMethod_;
  NSLPApp *nslp_;
  //GimpsTcpApp *agent_tcp_ ; 
  //sender and receiver tcp connections list
  GimpsTcpApp *sender_tcp_app;
  GimpsTcpApp *receiver_tcp_app;
  int transmissionMode;
  void prnt(char *method, char *msg);
  
protected:
  void reset_GIMPSAgent_Timer();
  GIMPSAgent_Timer GIMPS_timer_;							//timer for refresh message and send
  void send_message_pendent();
  void list_sessions_agent();
  double refresh_;  /* The default refresh value */
  int ip6_;     /* Should IPv6 message lengths be simulated? */
  int nam_;     /* Should the correct NSIS message names appear in nam? */
  int refreshment;  
  
private:
  void RoutingState_add(MessageRoutingState *mrs);
  void RoutingState_del(MessageRoutingState *mrs);
  void MessageAssociation_add(MessageAssociationState *ma);
  MessageAssociationState *getMessageAssociation(int p_Identity);
  void MessageAssociation_del(int p_Identity);
  TlvObjectQC *createQueryCookie();
  TlvObjectRC *createResponderCookie();
  TlvObjectNLI *createNLI(MessageRoutingState *mrs);
  TlvObjectSCD *createSCD();
  TlvObjectSP *createStackProposal();
  void processMessage(GIMPSMessage *msg);
  void forwardmessage(GIMPSMessage *msg);
  void add_message_for_agent(GIMPSMessage *ma);
  void sendQueryResponse(GIMPSMessage *oldmsg, MessageRoutingState *mrs_);
  GIMPSMessage *createResponseMessage(GIMPSMessage *oldmsg, 
              MessageRoutingState *mrs_, int conn_type);
  GIMPSMessage *createDataMessage(int nslpId, TlvObjectMRI *mri, 
			  TlvObjectSI *sessionId, char *nslpData);
  void sendQueryCookiebyMA(GIMPSMessage *oldmsg, MessageRoutingState *mrs_);
  void sendQueryConfirm(GIMPSMessage *oldmsg, MessageRoutingState *mrs_);
  GIMPSMessage *GIMPSAgent::createConfirmMessage(GIMPSMessage *oldmsg, 
                MessageRoutingState *mrs_,  int conn_type);
  void refresh_mrs_time(GIMPSMessage *msg);
  void addNSLPmessage(GIMPSMessage *msg);
  void createMessage(GIMPSMessage *oldmsg, char direction) ;
  MessageRoutingState *getRoutingState(TlvObjectMRI *mri, int nslpId,TlvObjectSI *si);
  MessageRoutingState *GIMPSAgent::getRoutingState2(TlvObjectMRI *mri, int nslpId, TlvObjectSI *si);
  char cmpMRS(MessageRoutingState *mrs1, MessageRoutingState *mrs2);
  GimpsTcpApp *createTcpConnection(nsaddr_t src, nsaddr_t dst);
  void setFid(int fid) { fid_ = fid; }
  void enqueueNslpMessage(NslpDataInfo *nslpinfo);
  void sendNslpQueuedMessages();
  int sessionIdcounter;
  int poolingTime;
  int debug;
  NslpDataInfo *nslp_data_info;

};

class NSLPmessage {
public:
	TlvObjectMRI *mri;
	TlvObjectSI *si;
	TlvObjectNSLP *NSLPData;
	char directionflow;
	int NslpID;
	nsaddr_t fromhop;
    NSLPmessage(TlvObjectMRI *mri , TlvObjectSI *si, TlvObjectNSLP *NSLPData, char directionflow, int NslpID, nsaddr_t fromnode);
    ~NSLPmessage(){ 
	   delete mri;
	   delete si;
	   if (NSLPData != NULL) {delete NSLPData;};
	};
	NSLPmessage *next;
private:
	
};

class TcpAppString : public AppData {
private:
	int size_;
	char* str_; 
public:
	TcpAppString() : AppData(TCPAPP_STRING), size_(0), str_(NULL) {}
	TcpAppString(TcpAppString& d) : AppData(d) {
		size_ = d.size_;
		if (size_ > 0) {
			str_ = new char[size_];
			strcpy(str_, d.str_);
		} else
			str_ = NULL;
	}
	virtual ~TcpAppString() { 
		if (str_ != NULL) 
			delete []str_; 
	}

	char* str() { return str_; }
	virtual int size() { return AppData::size() + size_; }

	// Insert string-contents into the ADU
	void set_string(char* s) {
		if ((s == NULL) || (*s == 0)) 
			str_ = NULL, size_ = 0;
		else {
			size_ = strlen(s) + 1;
			str_ = new char[size_];
			assert(str_ != NULL);
			strcpy(str_, s);
		}
	}
	virtual AppData* copy() {
		return new TcpAppString(*this);
	}
};

class NslpDataInfo {
public:
	NslpDataInfo(TlvObjectMRI *mri, char *data, TlvObjectSI *si);
	~NslpDataInfo();
	char *getData() { return data; }
	TlvObjectMRI *getMRI() { return mri; }
	TlvObjectSI *getSI() { return si; }
	TlvObjectMRI *mri;
	TlvObjectSI *si;
	char *data;
	NslpDataInfo *next;
};

class GimpsTcpApp: public TcpApp {
public:
	GimpsTcpApp(Agent *);
	virtual ~GimpsTcpApp();
	virtual void process_data(int size, AppData *data);
	void incRef() { num_ref++; }
	void decRef() { num_ref--; }
	void sendmsg(int nbytes, unsigned int data);
	GimpsTcpApp* getPeer() { return peer_; }
	int command(int argc, const char*const* argv);
	GIMPSAgent *getGimpsAgent() { return gimps_agent_; }
	GimpsTcpApp *next;
	GIMPSAgent *gimps_agent_;
	int num_ref;
	GimpsTcpApp *peer_;
	//virtual void send(int nbytes, AppData *data);
};

#endif

