#include "gimpsmessage.h"

void GIMPSMessage::setprotocol(char config)
{
  protocol = config;
};

char GIMPSMessage::getprotocol()
{
  return protocol;
};

char GIMPSMessage::receivebydatagram() {
  if (protocol == DATAGRAM)
     return 1;
  return 0; 
}

char GIMPSMessage::getgimpshops() 
{
  return header.gimpshops;
};  

GIMPSMessage::GIMPSMessage(int nslpId,  TlvObjectMRI *mri, 
 TlvObjectSI *sessionId, TlvObjectNSLP *nslpData)
{
  initialize() ;
  header.nslpId = nslpId;
  addTlvMRI(mri);
  addTlv(sessionId);
  if(nslpData != NULL)
  	addTlvNSLP(nslpData);
};

void GIMPSMessage::addTlv(TlvObject *obj)
{
  if (objects == NULL)
  {  
    objects = obj;
  }
  else
  {
    tail->next = obj;
  }
  obj->next = NULL;
  tail = obj;
  header.length = header.length + obj->get_length();
};

void GIMPSMessage::addTlvMRI(TlvObjectMRI *obj)
{
  this->MRI = obj;
  header.length = header.length + obj->get_length();
};

void GIMPSMessage::addTlvNSLP(TlvObjectNSLP *obj)
{ 
  this->NSLP = obj;
  if (obj != NULL) {
      header.length = header.length + obj->get_length();
  }
};

/* delete_objects() deletes all objects from the object list
 */
void GIMPSMessage::delete_objects()
{ 
  TlvObject *temp;
  while (objects != NULL)
  {
    temp = objects;
    objects = objects->next;
    
    delete temp;
  }
  delete NSLP;
  NSLP = NULL;
  delete MRI;
  MRI = NULL;
}

int GIMPSMessage::get_conlength(){
  return this->header.conlength;	
    	
};

int GIMPSMessage::get_length(){
   return this->header.length;	
    	
};

GIMPSMessage::GIMPSMessage(int nslpId, TlvObjectMRI *mri, 
TlvObjectSI *sessionId)
{
  initialize();
  header.nslpId = nslpId;
   // Config MRI
  addTlvMRI(mri);
   // Config SESSION ID
  addTlv(sessionId);
};

TlvObject *GIMPSMessage::findTlv(int classnum)
{
  TlvObject *search;
  search = objects;
  while (search != NULL) 
  {
    if (search->getTlvObjectType() == classnum) {	  
  	  return search;
     }	  
     search = search->next;
  };
  
  return search;
};

TlvObjectSI *GIMPSMessage::getTlvSI()
{
 /*
  TlvObjectSI *tmp2;
 
  TlvObject *tmp = findTlv(TYPE_SESSION_IDENTIFICATION);
 
  if (tmp != NULL) {
 
     tmp2 = new TlvObjectSI(tmp->get_contents());
    } else {  
     printf("\n gimpmessage while i was running gettlvsi: I didnt find any object Session \n"); 
     exit(0);
   };
 
  return tmp2;
  */
  TlvObjectSI *tmp2;
  TlvObject *tmp = findTlv(TYPE_SESSION_IDENTIFICATION);
  if (tmp == NULL) {  
     printf("\n gimpmessage while i was running getTlvSi: I didnt find any object Session \n"); 
     exit(0);
  }
  tmp2 = (TlvObjectSI *)tmp;
  return tmp2;
};

TlvObjectQC *GIMPSMessage::getTlvQC()
{
  TlvObject *tmp = findTlv(TYPE_QUERY_COOKIE);
  if (tmp == NULL) {
     printf("[getTlvQC]: The object Query Cookie is not included in this message\n"); 
     exit(0); 	
	 return NULL;  
  }  
  TlvObjectQC *tmp2 = new TlvObjectQC(tmp->get_contents());
  return tmp2;
};

TlvObjectRC *GIMPSMessage::getTlvRC()
{
  TlvObject *tmp = findTlv(TYPE_RESPONDER_COOKIE);
    if (tmp == NULL) {
     printf("[getTlvQC]: The object Response Cookie is not include in this message\n"); 
     return NULL; 	
  }  
  TlvObjectRC *tmp2 = new TlvObjectRC(tmp->get_contents());
  return tmp2;
};

int GIMPSMessage::getPeerIdentity()
{
  int c;
  TlvObject *tmp = findTlv(TYPE_NETWORK_INFORMATION_LAYER);
  TlvObjectNLI *tmp2 = new TlvObjectNLI(tmp->get_contents());
  c=tmp2->getPeerIdentity();
  delete tmp2;
  return c;
};


TlvObjectNLI  *GIMPSMessage::getTlvNLI()
{
  /*
  TlvObjectNLI *tmp2;
  TlvObject *tmp = findTlv(TYPE_NETWORK_INFORMATION_LAYER);
  if (tmp != NULL)
  {
    tmp2 = new TlvObjectNLI(tmp->get_contents());
    return tmp2;
  }
  return NULL; 
  */
  TlvObjectNLI *tmp2;
  TlvObject *tmp = findTlv(TYPE_NETWORK_INFORMATION_LAYER);
  if (tmp != NULL)
  {
    tmp2 = (TlvObjectNLI*)tmp;
    return tmp2;
  }
  return NULL;
};

 
int  GIMPSMessage::getRSVT()
{ int time;
  TlvObjectNLI *tmp2;
  TlvObject *tmp = findTlv(TYPE_NETWORK_INFORMATION_LAYER);
  if (tmp != NULL)
  {
    tmp2 = new TlvObjectNLI(tmp->get_contents());
    time = tmp2->getRSVT();
    delete tmp2;
    return time;
  }
  return -1;
};

TlvObjectSCD *GIMPSMessage::getTlvSCD()
{
  TlvObjectSCD *tmp2;
  TlvObject *tmp = findTlv(TYPE_SCD);
  if (tmp != NULL)
  {
    tmp2 = new TlvObjectSCD(tmp->get_contents());
    return tmp2;
  }
  return NULL;
};

TlvObjectSP *GIMPSMessage::getTlvSP()
{
  TlvObjectSP *tmp2;
  TlvObject *tmp = findTlv(TYPE_STACK_PROPOSAL);
  if (tmp != NULL)
  {
    tmp2 = new TlvObjectSP(tmp->get_contents());
    return tmp2;
  }
  return NULL;
};

TlvObjectMRI *GIMPSMessage::getTlvMRI()
{ 
  /*
  if (MRI == NULL) {
     printf("--Object MRI it is not present in the message, you probably have forgot this include");
     exit(0);
     return NULL;
  }
  TlvObjectMRI *tmp2 = new TlvObjectMRI(MRI->get_contents());
  return tmp2;
  */
  if (MRI == NULL) {
     printf("--Object MRI it is not present in the message, you probably have forgot this include");
     exit(0);
     return NULL;
  }
  return MRI;
};

TlvObjectNSLP *GIMPSMessage::getTlvNSLP()
{
  TlvObjectNSLP *tmp = NULL;
   if (NSLP != NULL)
  {
   tmp = new TlvObjectNSLP(NSLP->getNSLP());
   };

  return tmp;
};

nsaddr_t GIMPSMessage:: getDestinationAddress()
{
  return destinationaddr;
};

int GIMPSMessage::getDestinationPort()
{
  return  destinationport;
};

nsaddr_t GIMPSMessage::getSourceAddress()
{
  return  MRI->getSourceAddress();
};

int GIMPSMessage::getSourcePort()
{
  return MRI->getSourcePort();
};

int GIMPSMessage::getDirection()
{
  return MRI->getDirection();
};

int GIMPSMessage::getNslpId()
{
  return header.nslpId;
};

char GIMPSMessage::decreaseHopCounter()
{
  header.gimpshops--;
  if(header.gimpshops <= 0) return 0;
  return 1;
};

char GIMPSMessage::hasQuery()
{
  TlvObject *tmp = findTlv(TYPE_QUERY_COOKIE);
  if (tmp==NULL) return 0;
  return 1;
};

char GIMPSMessage::hasResponder()
{
  TlvObject *tmp = findTlv(TYPE_RESPONDER_COOKIE);
  if (tmp==NULL) return 0;
  return 1;
};

char GIMPSMessage::hasNslp()
{
 
  if (NSLP == NULL) return 0;
  return 1;
};

char GIMPSMessage::hasNLI()
{
  TlvObject *tmp = findTlv(TYPE_NETWORK_INFORMATION_LAYER);
  if (tmp==NULL) return 0;
  return 1;
};

void GIMPSMessage::configHopCount(int num)
{
  header.gimpshops = num;
};

void GIMPSMessage::initialize()
{
 //header.type = 0;
  settypemsg(0);	       					// The message type
  header.send_TTL = 10;   					// The IP TTL with which the message was sent
  header.length = 64;      					// The "simulated" length of the message
  header.conlength = 0;   					// The "real" length of the message
  header.headerSize = 64;
  header.version = 0;
  header.gimpshops = 0;
  header.messageLength = 0;
  header.nslpId = 0;
  header.direction = 0;
  header.reserved = 11;
  protocol = 0;
  objects = NULL;
  tail = NULL ;
  MRI = NULL;
  NSLP = NULL;
  next = NULL;
  prev = NULL;
  forward = 0;
  fromhop = -1;
  receivebyUDP = 0;
};

MessageRoutingState::MessageRoutingState()
{
  inicialization();
}

void MessageRoutingState::inicialization()
{
  active = 0;
  maxRetry = 2;					// Max retry for the peers timers
  timeOutTime = 30;// Default time out time: 30 seconds (30000 milliseconds)
  downstreamTimerValue = -1;
  downstreamRetry = 0;
  downstreamPeer = -1;
  upstreamPeer = -1;
  upstreamPeerMASPointer = NULL;
  downstreamPeerMASPointer = NULL;
  IPTTL=0;
 
};

MessageRoutingState::MessageRoutingState(GIMPSMessage *msg1, 
 MessageAssociationState *peerMASPointer_, int handler, char active_)
{ //createsa MRS with MA
  TlvObjectNLI  *nli;
  inicialization(); 
  nli = new TlvObjectNLI(msg1->getTlvNLI()->get_contents());
  this->mri = new TlvObjectMRI(msg1->getTlvMRI()->get_contents());  // Primary Key
  this->sessionId = new TlvObjectSI(msg1->getTlvSI()->get_contents());
  this->nslpId = msg1->getNslpId();
  this->nslpMessageHandle = handler;
  this->mri->changeDirection();
  if(nli != NULL) 
  { 
    this->IPTTL = nli->getIPTTL();
    this->timeOutTime = nli->getRSVT();
    //this->IPTTL = msg1->IPTTL;
  }
  if(this->mri->getDirection() == UPSTREAM)
  { //creating a MRS at the receiver GIMPS node
  	this->upstreamPeerMASPointer = peerMASPointer_;
    this->upstreamTimerValue = this->timeOutTime;
    this->upstreamTimer = this->timeOutTime;
  }  
  else {
  	//TODO: to implement the MASPointer at the Sender side
  	this->downstreamPeerMASPointer = peerMASPointer_;
    this->downstreamTimerValue = this->timeOutTime;
    this->downstreamTimer = this->timeOutTime;
  }
  delete nli;
  this->active = active_;
}

MessageRoutingState::MessageRoutingState(GIMPSMessage *msg1, int handler,
 char active_)
{ // setup a MRS without a MA
  TlvObjectNLI  *nli;
  inicialization();
  nli = new TlvObjectNLI(msg1->getTlvNLI()->get_contents());
  //this->IPTTL = nli->getIPTTL(); seg.fault
  //this->IPTTL = msg1->IPTTL;
  this->mri = new TlvObjectMRI(msg1->getTlvMRI()->get_contents()); 				// Primary Key
  this->sessionId = new TlvObjectSI(msg1->getTlvSI()->get_contents());
  this->nslpId = msg1->getNslpId();
  this->nslpMessageHandle = handler;  			// NSLP Handle
  this->mri->changeDirection();
  if(this->mri->getDirection() == UPSTREAM)
  { //createing a MRS at the receiver GIMPS node
    if(nli != NULL) 
    { 
      this->IPTTL = nli->getIPTTL();
      this->upstreamPeer = nli->getInterfaceAddress();
      this->upstreamPeerPort = 0;      //jsut simulation is not necessary  upstream its made by UDP;
      this->upstreamTimerValue = this->timeOutTime;
      this->upstreamTimer = this->timeOutTime;
      //printf("[MessageRoutingState constructor] upstreamTimer: %d at Node \n", this->upstreamTimer);    
    }
    else
    {
      //TODO: 
    }
  }
  else
  { //creating a MRS at the sender GIMPS node
    if(nli != NULL) 
    {
      	this->IPTTL = nli->getIPTTL();
      	this->downstreamPeer = nli->getInterfaceAddress();
      	this->downstreamPeerPort = 0;      //jsut simulation is not necessary  upstream its made by UDP;
      	this->downstreamTimerValue = this->timeOutTime;
      	this->downstreamTimer = this->timeOutTime;
      	//printf("[MessageRoutingState constructor] downstreamTimer: %d at Node \n", this->downstreamTimer);
    }	
  }
  delete nli;
  this->active = active_;		// Set active or inactive
};

//long MessageRoutingState::MessageRoutingState::getDownstreamTimer()
long MessageRoutingState::getDownstreamTimer()
{
  return downstreamTimer;
};

void MessageRoutingState::resetDownstreamTimer()
{
  downstreamTimer = downstreamTimerValue;
};

long MessageRoutingState::getUpstreamTimer()
{
  return upstreamTimer;
};

void MessageRoutingState::resetUpstreamTimer()
{
  upstreamTimer = upstreamTimerValue;
};

int MessageRoutingState::getDownstreamRetry()
{
  return downstreamRetry;
};

int MessageRoutingState::resetDownstreamRetry()
{
  downstreamRetry = 0;
};

int MessageRoutingState::getNslpMessageHandle()
{
  return nslpMessageHandle;
};

void MessageRoutingState::addDownstreamRetry()
{
  //resetDownstreamTimer();
  downstreamRetry++;
};

void MessageRoutingState::decreaseDownstreamTimer(int time)
{
  //the timer of the sender is decreased faster than the one of the receiver
  downstreamTimer -= time; 

  if (downstreamTimer < 0) downstreamTimer = 0;
};

void MessageRoutingState::decreaseUpstreamTimer(int time)
{
  upstreamTimer -= time;

  if (upstreamTimer < 0) upstreamTimer = 0;
};

int MessageRoutingState::getNslpId()
{
  return nslpId;
};

TlvObjectMRI *MessageRoutingState::getMri()
{
  //TlvObjectMRI *tmp2 = new TlvObjectMRI(mri->get_contents());
  return mri;
};

char MessageRoutingState::hasDownstreamPeerInfo()
{
  if ((downstreamPeerMASPointer != NULL) || (downstreamPeer != -1))	
  	return 1;
  return 0;
};

char MessageRoutingState::hasUpstreamPeerInfo()
{
  if ((upstreamPeerMASPointer != NULL) || (upstreamPeer != -1))
    return 1;
  return 0;
};

MessageAssociationState *MessageRoutingState::getUpstreamPeerMA()
{
  return upstreamPeerMASPointer;
};

MessageAssociationState *MessageRoutingState::getDownstreamPeerMA()
{
  return downstreamPeerMASPointer;
};

nsaddr_t MessageRoutingState::getUpstreamPeer()
{
  return upstreamPeer;
};

nsaddr_t MessageRoutingState::getDownstreamPeer()
{
  return downstreamPeer;
};

TlvObjectSI *MessageRoutingState::getSessionId()
{
  //TlvObjectSI *tmp2 = new TlvObjectSI(sessionId->get_contents());
  //return tmp2;
  return sessionId;
};

void MessageRoutingState::setUpstreamPeer(nsaddr_t upstreamPeer, nsaddr_t upstreamPeerPort, long timeOut_)
{
  // Set time ou time
  this->timeOutTime = timeOut_;

  // Update Upstream Peer
  this->upstreamPeer = upstreamPeer;
  this->upstreamPeerPort = upstreamPeerPort;
  this->upstreamTimerValue = this->timeOutTime;
  this->upstreamTimer = this->timeOutTime;
  this->upstreamPeerMASPointer = NULL;
};

void MessageRoutingState::setUpstreamPeer(MessageAssociationState * upstreamPeerMASPointer, long timeOut_)
{
  // Set time ou time
  this->timeOutTime = timeOut_;
  this->upstreamPeerMASPointer = upstreamPeerMASPointer;
  this->upstreamTimerValue = this->timeOutTime;
  this->upstreamTimer = this->timeOutTime;
};

void   MessageRoutingState::setDownstreamPeer(MessageAssociationState *ma_, long timeOut_)
{
  this->downstreamPeerMASPointer = ma_;
  this->downstreamTimerValue = this->timeOutTime;
  this->downstreamTimer = this->timeOutTime;
  this->active = 1;
};

void MessageRoutingState::deleteDownstreamInfo()
{
  this->downstreamPeerMASPointer = NULL;
  this->downstreamPeer = -1;
  this->downstreamPeerPort = -1;
};

void MessageRoutingState::deleteUpstreamInfo()
{
  this->upstreamPeerMASPointer = NULL;
  this->upstreamPeer = -1;
  this->upstreamPeerPort = -1;
};

//GimpsTcpApp *MessageRoutingState::getAssociation() {
TcpApp *MessageRoutingState::getAssociation() {
	MessageAssociationState *ma;
	
	ma = getDownstreamPeerMA();
	if (ma == NULL) {
		ma = getUpstreamPeerMA();
		if(ma == NULL)
			return NULL;
	}
	return ma->getAssociation();	
}
//void MessageRoutingState::setAssociation(GimpsTcpApp *tcp_agent) {
void MessageRoutingState::setAssociation(TcpApp *tcp_agent) {
	MessageAssociationState *ma;
	
	ma = getDownstreamPeerMA();
	if (ma == NULL) {
		ma = getUpstreamPeerMA();
		if(ma == NULL)
			return;
	}
	//tcp_agent->incRef();
	ma->setAssociation(tcp_agent);
}

void MessageRoutingState::delAssociation() {
	MessageAssociationState *ma;
	
	ma = getDownstreamPeerMA();
	if (ma == NULL) {
		ma = getUpstreamPeerMA();
		if(ma == NULL)
			return;
	}
	ma->delAssociation();
}

//-------------------------------------------------------------- Message associate ---------------

void MessageAssociationState::delAssociation() {
	// ??
	if (association == NULL)
		return;
	//association->decRef();
	//delete(association);
	association = NULL;		
}

MessageAssociationState::MessageAssociationState(int p_identity, 
nsaddr_t p_interface, int p_port, long p_messageTimer)
{
  peer_interface = p_interface;
  peerport = p_port;
  peerIdentity = p_identity;
  peermessageTimer = p_messageTimer;
  active = 1;
  association = NULL;
};

MessageAssociationState::MessageAssociationState(TlvObjectNLI *nli_, 
long p_messageTimer)
{
  peer_interface = nli_->getInterfaceAddress();
  peerport = 0;					//we are just simulated that we are send for some ports.
  peerIdentity = nli_->getPeerIdentity();
  peermessageTimer = p_messageTimer;
  active = 1;
  association = NULL;
};