#ifndef ns_nsis_messages_h
#define ns_nsis_messages_h

#include "nsis-objects.h"
#include "timer-handler.h"
#include "random.h"
#include "webcache/tcpapp.h"
//#include "gimps.h"
//#include "gimps-tcp-app.h"

#define CURRENT_TIME  Scheduler::instance().clock()
#define JITTER	 (Random::uniform()*0.5)

class GIMPSMessage
{
public :
  int DIRECTION_UPSTREAM ;
  int DIRECTION_DOWNSTREAM ;
  ~GIMPSMessage(){ delete_objects(); };
  GIMPSMessage(int nslpId,  TlvObjectMRI *mri, TlvObjectSI *sessionId, TlvObjectNSLP *nslpData);
  GIMPSMessage(int nslpId,  TlvObjectMRI *mri, TlvObjectSI *sessionId);
  void addTlv(TlvObject *obj) ;
  void addTlvMRI(TlvObjectMRI *obj);
  void addTlvNSLP(TlvObjectNSLP *obj);
  TlvObject *findTlv(int classnum) ;
  nsaddr_t getDestinationAddress();
  int getDestinationPort() ;
  nsaddr_t getSourceAddress();
  inline void setdestinationaddr(nsaddr_t config) {destinationaddr = config;};
  inline void setdestinationport(int config) {destinationport = config;};
  int getSourcePort() ;
  int getDirection();
  int getNslpId() ;
  char decreaseHopCounter(); 
   /*  switch (msg->get_type()) {
      case PATH: process_path_message(msg, cmn_hdr->iface_, ret); break;
      case PATHTEAR: process_path_tear_message(msg, cmn_hdr->iface_, ret); break;
      case RESV: process_resv_message(msg, rsvp_hdr->fromhop); break;
      case RESVTEAR: process_resv_tear_message(msg); break;
      case RESVERR: process_resv_err_message(msg); break;
      case RESVCONF: process_resv_conf_message(msg); break;
    } */
  char hasQuery();
  char hasResponder();
  char hasNslp();
  char hasNLI();
  void configHopCount(int num);
  void delete_objects();
  void setprotocol(char config);
  inline void settypemsg(int config) {header.type = config;}
  inline int gettypemsg() {return header.type;}
  char getprotocol();
  TlvObjectSI   *getTlvSI();
  TlvObjectQC   *getTlvQC();
  TlvObjectRC   *getTlvRC();
  TlvObjectMRI  *getTlvMRI();
  TlvObjectNLI  *getTlvNLI();
  TlvObjectSCD  *getTlvSCD();
  TlvObjectSP   *getTlvSP();
  TlvObjectNSLP *getTlvNSLP();
  int getRSVT();
  int getPeerIdentity();
  int get_conlength();  
  int get_length();  
  char receivebydatagram();
  char getgimpshops();
  char protocol;
  
  TlvObjectMRI *MRI;
  GIMPSMessage *next, *prev;
  char forward;
  nsaddr_t fromhop;
  struct common_header
  {
    int type;       					// The message type
    char send_TTL;   					// The IP TTL with which the message was sent
    int length;      					// The "simulated" length of the message
    int conlength;   					// The "real" length of the message
    int headerSize ;
    int version ;
    int gimpshops ;
    int messageLength ;
    int nslpId ;
    int direction ;
    int reserved ;
  };
  char receivebyUDP ;
  struct common_header header;

 private:
  
  void initialize();

  // The common header for all NSIS messages

  // GIMPS TLV
  TlvObject *objects ;
  TlvObject *tail ;

  TlvObjectNSLP *NSLP;
  nsaddr_t destinationaddr;
  int      destinationport;
};
 
class MessageAssociationState
{
public:
  long messageTimer;
  ~MessageAssociationState() {  }; 
  MessageAssociationState(int p_identity, nsaddr_t p_interface, int p_port, long p_messageTimer);
  MessageAssociationState(TlvObjectNLI *na_, long p_messageTimer);
  MessageAssociationState(MessageAssociationState *ma_, long p_messageTimer);
  inline nsaddr_t getPeerInterface()    {return peer_interface;};
  inline int      getPort()            {return peerport;};
  int getPeerIdentity() {return peerIdentity;};   
  inline void setmessagetimer(int config) {peermessageTimer = config;}
  inline int  getmessagetimer() {return peermessageTimer;}
  //GimpsTcpApp *getAssociation() {return association;}
  TcpApp *getAssociation() {return association;}
  //void setAssociation(GimpsTcpApp *tcp_agent) {association = tcp_agent;}
  void setAssociation(TcpApp *tcp_agent) {association = tcp_agent;}
  void delAssociation();
  MessageAssociationState *next,
                          *prev;
  
private:
  nsaddr_t peer_interface;
  int peerport;
  int peerIdentity;  
  int peermessageTimer;
  char active;
  //GimpsTcpApp *association;
  TcpApp *association;
 };

class MessageRoutingState
{
public:// State of this MRS. If false means that it was send a query and it's waiting for the responder
  ~MessageRoutingState(){ 
     delete mri;
     delete sessionId;
     };
  MessageRoutingState();
  MessageRoutingState(GIMPSMessage *msg1, MessageAssociationState *upstreamPeerMASPointer_, int handler, char active_);
  MessageRoutingState(GIMPSMessage *msg1, int handler, char active_); 

  long    getDownstreamTimer();
  long    getUpstreamTimer();
  void    resetDownstreamTimer();
  void    resetUpstreamTimer();
  int     getDownstreamRetry();
  int     resetDownstreamRetry();
  int     getNslpMessageHandle();
  void    addDownstreamRetry();
  void    decreaseDownstreamTimer(int time);
  void    decreaseUpstreamTimer(int time) ;
  int     getNslpId();
  TlvObjectMRI  *getMri();
  char    hasDownstreamPeerInfo();
  char    hasUpstreamPeerInfo();
  nsaddr_t getUpstreamPeer();
  nsaddr_t getDownstreamPeer();
  MessageAssociationState *getUpstreamPeerMA();
  MessageAssociationState *getDownstreamPeerMA();
  TlvObjectSI   *getSessionId();
  TlvObjectNSLP *getTlvNSLP();
  void    setUpstreamPeer(nsaddr_t upstreamPeer, nsaddr_t upstreamPeerPort, long timeoOut_);
  void    setUpstreamPeer(MessageAssociationState *upstreamPeerMASPointer, long timeOut_);
  void    setDownstreamPeer(MessageAssociationState *ma_, long timeOut_);
  inline void    setupstreamtime(long config) {upstreamTimer = config;};  
  inline void    setdownstreamtime(long config) {downstreamTimer = config;};
  void    deleteDownstreamInfo();
  void    deleteUpstreamInfo();
  //GimpsTcpApp *getAssociation();
  TcpApp *getAssociation();
  void setAssociation(TcpApp *);
  void delAssociation();
  int 	active ;
  int 	maxRetry ;					// Max retry for the peers timers
  long timeOutTime ;				// Default time out time: 30 seconds (30000 milliseconds)
  MessageRoutingState *next;
  MessageAssociationState *upstreamPeerMASPointer;
  MessageAssociationState *downstreamPeerMASPointer;
  TlvObjectSI *sessionId;  
  TlvObjectMRI *mri;
  int nslpId;
  int IPTTL; 
private:
  void inicialization();
 
 
  // To prevent upstream for a given MRI being corrupted by a malicious upstream node.
 

  /*
   * Upstream peer: MessageAssociationStatethe adjacent peer closer to the flow source.  This
   *   could be an IP address and UDP port (learned from previous
   *   signaling) or a pointer to a valid messaging association.  It
   *   could also be null, if this node is not storing reverse routing
   *   state or if it is the last upstream node (including the sender).
   */
  nsaddr_t upstreamPeer;
  int upstreamPeerPort;
  long upstreamTimerValue;
  long upstreamTimer;

  /*
   * Downstream peer: the adjacent peer closer to the flow destination.
   *   This could be a pointer to a valid messaging association, or it
   *   could be null, if this node is only sending downstream datagram
   *   mode messages for this flow and signaling application, or if it is
   *   the last downstream node (including the receiver).
   */
  nsaddr_t downstreamPeer;
  int downstreamPeerPort;
  long downstreamTimerValue;
  long downstreamTimer;
  int downstreamRetry ;
  int nslpMessageHandle;// Handle to the NSLP responsible for this MRS (TODO: needed?)
};




#endif


