/*
 * Copyright (c) 1998 The University of Bonn
 * All rights reserved.
 * 
 * Permission to use and copy this software in source and binary forms
 * is hereby granted, provided that the above copyright notice, this
 * paragraph and the following disclaimer are retained in any copies
 * of any part of this software and that the University of Bonn is
 * acknowledged in all documentation pertaining to any such copy
 * or derivative work. The name of the University of Bonn may not
 * be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL 
 * THE UNIVERSITY OF BONN BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF 
 * OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE.
 */

#include "nsis-link.h"


static class NSISLinkClass : public TclClass
{
public:
  NSISLinkClass() : TclClass("NSISLink") {}
  TclObject* create(int, const char*const*)
  {
    return (new NSISLink());
  }
}
class_NSISLink;


static class NSISCheckerClass : public TclClass
{
public:
  NSISCheckerClass() : TclClass("NSISChecker") {}
  TclObject* create(int, const char*const*)
  {
    return (new NSISChecker());
  }
}
class_NSISChecker;


NSISChecker::NSISChecker()
{
  //bind("off_ip_", &off_ip_);
  //bind("off_nsis_", &off_nsis_);
  bind("src_", &src_);
}


NSISLink::NSISLink()
{
  bind_bw("besteffort_", &besteffort_);
}


/* The 'give' function allows the NSIS agent to hand a packet back
   to the link after it was processed. That way, NSIS agents don't have
   to worry about routing (especially multicast routing). */
void NSISChecker::give(Packet *p)
{
  /* Simply forward the packet as if there was no NSIS agent */
  target_->recv(p);
}


void NSISChecker::recv(Packet* p, Handler* h)
{ char ret=0;
  
  int temp;
  //char *cc = new char[3000];  
  hdr_cmn *hdr = hdr_cmn::access(p);
  hdr->last_hop_ = src_;
  hdr_ip *ip_hdr = hdr_ip::access(p);
  //printf("\n[NSISChecker] --recv()--\n");
  /*
  if ((hdr->ptype() == PT_NSIS)) printf("[NSISChecker]protocol NSIS\n");
  if ((hdr->ptype() == PT_TCP)) printf("[NSISChecker]protocol tcp\n");
  if ((hdr->ptype() == PT_ACK)) printf("[NSISChecker]protocol ACK\n");  
 */
  temp = 0;
  if ((hdr->ptype() == PT_NSIS))
  {
    Tcl& tcl = Tcl::instance();
    tcl.evalf("%s get-gimps-agent", name());
    if (strlen(tcl.result()) == 0)
    {
      printf("[NSISChecker recv] Destination agent not found!\n");	
      target_->recv(p);
      return;
    }
    if (ip_hdr->daddr() == ((GIMPSAgent *)TclObject::lookup(tcl.result()))->gimpsIP)
    { 
      //The message will correctly be delivered to GIMPS Agent by the respective node
      target_->recv(p);
      return;
    }
    //printf("[NSISChecker - recv] Packet destination: %d GimpsIP: %d\n", 
    //ip_hdr->daddr(), ((GIMPSAgent *)TclObject::lookup(tcl.result()))->gimpsIP); 
    hdr_nsis *nsishdr = hdr_nsis::access(p);
    nsishdr->fromhop = src_;
    ret = ((GIMPSAgent *)TclObject::lookup(tcl.result()))->give(p, this);
    if (ret == 0) {
    	// the packet will be not forwarded.
    	// It is redirected to the same node to be destructed.
    	ip_hdr->prio_ = -1;
    	ip_hdr->daddr() = ((GIMPSAgent *)TclObject::lookup(tcl.result()))->gimpsIP;
    }
    target_->recv(p);
    return;	      
  }
  target_->recv(p);
}
