
/***************************************************************************
 *   Copyright (C) 2005 by Ribeiro, Romulo Eugenio                         *
 *   romulo@dei.uc.pt                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ns_nsis_objects_h
#define ns_nsis_objects_h


#include "config.h"

//#include <stdio.h>
//#include <string.h>
//#include <iostream.h>


#define TYPE_MESSAGE_ROUTING_INFORMATION	 1
#define TYPE_SESSION_IDENTIFICATION 	2
#define TYPE_NETWORK_INFORMATION_LAYER 	3
#define TYPE_SCD			11

#define TYPE_STACK_PROPOSAL 		4
#define TYPE_QUERY_COOKIE 		5
#define TYPE_RESPONDER_COOKIE 		6
#define TYPE_LIFETIME 			7
#define TYPE_NAT_TRAVERSAL 		8
#define TYPE_NSLP_DATA 			9
#define TYPE_ERROR 			10
#define DATAGRAM 		          1
#define CONNECTION 		          2
#define IPV4 			          1
#define IPV6 			          2
#define SIZECOOKIE 		         16
#define UPSTREAM   			  0
#define DOWNSTREAM			  1

#define MRI_PATH_COUPLED 1

#define ERROR_CLASS_INFORMATIONAL 	   1
#define ERROR_CLASS_SUCCESS 	   2
#define ERROR_CLASS_PROTOCOL_ERROR    3
#define ERROR_CLASS_TRANSIENT_FAILURE 4
#define ERROR_CLASS_PERMANENT_FAILURE 5
#define TYPE_QUERY_MSG 1
#define TYPE_RESPONSE_MSG 2
#define TYPE_CONFIRM_MSG 3
#define TYPE_DATA_MSG 4
#define TYPE_HELLO_MSG 5
#define CMODE 0
#define DMODE 1

#define NSISMESSAGE                   1001
// typedef int  nsaddr_t ;                              just for test

typedef char CHAR16[16];
typedef char CHAR32[32];


/* The common header for all NSIS objects */
struct objheader
{
  int length;     // The "simulated" length of the object
  int conlength;  // The "real" length of the object contents
  char aFlag;
  char bFlag;
  int typeFlags;
  int type;        // type of object
  int lengthFlags;
};

struct List_HI
{      //High layer protocol definition
  char protocol;
  int  port;
  long HIdata;
};

class TlvObject
{
public:
  TlvObject() : next(0) {};
  virtual ~TlvObject() { delete[] contents; };
  inline int get_length() { return header->length; };
  inline void set_length(int len) { header->length = len; };
  inline int get_conlength() { return header->conlength; };
  inline unsigned char *get_contents() { return contents; };
  inline int getTlvObjectType() {return header->type;};
  inline void configBFlag(char config) {header->bFlag = config;};
  inline void configAFlag(char config) {header->aFlag = config;};
  inline void configType(int config)   {header->type  = config;};
  TlvObject *next;
protected:
  struct objheader *header;
  unsigned char *contents;  // A pointer to the contents
};

class TlvObjectSI : public TlvObject
{
public:

  TlvObjectSI(int config, int fid_) ;
  TlvObjectSI(unsigned char *cont);
  virtual ~TlvObjectSI() { };
  inline void configSessionId (int config) {con->sessionId = config;};
  inline int getSessionId() {return con->sessionId;};
  inline int get_fid() { return con->fid; };

  void dump_object();

private:
  struct construct
  {
    int sessionId;
    int fid;
  };
  struct construct *con;
};
//-------------------------------------------------- qc -----------------

class TlvObjectQC: public TlvObject
{
public:
  TlvObjectQC() ;
  TlvObjectQC(unsigned char *cont);
  virtual ~TlvObjectQC() {};
  char *getQuery() ;
  void Generatecookie();

private:
  struct construct
  {
    //CHAR16 query;
    char query[16];
  };
  struct construct *con;
};
//-------------------------------------------------- rc ------------------------------------------------------

class TlvObjectRC: public TlvObject
{
public:
  TlvObjectRC();
  TlvObjectRC(unsigned char *cont);
  virtual ~TlvObjectRC() {} ;
  char *getResponder() ;
  void Generatecookie();
private:
  struct construct
  {
    //CHAR16 responder;
    char responder[16];
  };
  struct construct *con;
};
//-------------------------------------------------- lt -------------------------------------------------------

class TlvObjectLT: public TlvObject
{
public:
  TlvObjectLT(long config) ;
  TlvObjectLT(unsigned char *cont);
  virtual ~TlvObjectLT() {} ;
  inline void configLT(int config) { con->lt = config;};
  inline long getLT() {return con->lt;} ;
private:
  struct construct
  {
    long lt;
  };
  struct construct *con;
};
//-------------------------------------------------- MRI -------------------------------------------------------

class TlvObjectMRI: public TlvObject
{
public:
  virtual ~TlvObjectMRI(){ };
  TlvObjectMRI(unsigned char *cont);
  //d,a,and b flags are not in a correct order. The right order would be: a, b, d. 
  TlvObjectMRI(int messageRoutingMethod, char ipVer, char pFlag, char tFlag, char fFlag, char sFlag, char dFlag, char aFlag, char bFlag, int reservedFlag, nsaddr_t sourceAddress, nsaddr_t destinationAddress, int sourcePrefix, int destinationPrefix, char protocol,  int trafficClass, int reservedFlowLabel, int flowLabel, int spi, int sourcePort, int destinationPort); 
  
  inline void configMessageRoutingMethod(int config) { con->messageRoutingMethod = config;};
  inline void configIpVer(int config) 		{con->ipVer = config;};
  inline void configSourceAddress(int config) 	{con->sourceAddress = config;};
  inline void configDestinationAddress(int config) { con->destinationAddress = config;};
  inline nsaddr_t getDestinationAddress() 	{return con->destinationAddress;};
  inline nsaddr_t getSourceAddress() 		{return con->sourceAddress;};
  inline void configSourcePrefix(int config) 	{ con->sourcePrefix = config;};
  inline void configDestinationAddressPrefix(int config) { con->destinationPrefix =config;};
  inline void configProtocol(int config) 		{con->protocol = config;} ;
  inline void configTrafficClass(int config) 	{con->trafficClass = config;};
  inline int getDestinationPort() 			{return con->destinationPort;};
  inline int getSourcePort() 			{return con->sourcePort;};
  inline int getFlowLabel() 			{return con->flowLabel;};
  inline int getIpVer() 				{return con->ipVer;};
  inline int getProtocol() 				{return con->protocol;}
  inline int getSourceAddressPrefix() 		{return con->sourcePrefix ;};
  inline int getMessageRoutingMethod() 		{return con->messageRoutingMethod;};
  inline int getSpi() 				{return con->spi ;};
  inline int getTrafficClass() 			{return con->trafficClass;};
  int size();
  void configFlowLabel(int config) ;
  void configSpi(int config);
  void configSourcePort(int config);
  void configDestinationPort(int config);
  char changeDirection();
  char getDirection();
  void configDirection(char config);

private:
  struct construct
  {
    int messageRoutingMethod ;
    // Path-coupled
    char ipVer ;
    char pFlag ;
    char tFlag ;
    char fFlag ;
    char sFlag ;
    char dFlag ;
    char aFlag ;
    char bFlag ;
    int reservedFlag  ;
    nsaddr_t sourceAddress;
    nsaddr_t destinationAddress;
    int sourcePrefix ;
    int destinationPrefix ;
    int protocol ;
    int trafficClass ;
    int reservedFlowLabel; 					// Optional
    int flowLabel ; 						// Optional
    int spi ; 							// Optional
    int sourcePort ; 						// Optional
    int destinationPort ; 					// Optional
  } ;
  struct construct *con;
};
//------------------------------------------------------- SP ------------------------------------------------------

class TlvObjectSP:  public TlvObject
{
public:

  virtual ~TlvObjectSP(){};
  TlvObjectSP(unsigned char *cont);
  TlvObjectSP(int profCount, CHAR32 profiles);

private:

  struct construct
  {
    int profCount ;
    int reserved ;
    CHAR32 profiles;
  };

  struct construct *con;
};
// -------------------------------------------------- object Na -------------------------------------------

class TlvObjectNLI: public TlvObject
{
public:
  virtual ~TlvObjectNLI(){};
  TlvObjectNLI(unsigned char *cont);
  TlvObjectNLI(int piLength, int ipTTL, long RSVT, int ipVer, int peerIdentity, nsaddr_t interfaceAddress);
  inline void configPiLength(int config) 			{ con->piLength = config;};
  inline void configipTTL(int config) 			{ con->ipTTL = config;};
  inline void configIpVer(int config) 			{ con->ipVer = config;};
  inline void configInterfaceAddress(nsaddr_t config) 	{ con->interfaceAddress = config;};
  inline int getPeerIdentity()				{return con->peerIdentity;};
  inline nsaddr_t getInterfaceAddress()			{return con->interfaceAddress;};
  void configPeerIdentity(int config) 			{con->peerIdentity = config;};
  inline void configRSVT(int config) { con->RSVT = config;};
  inline long getRSVT() {return con->RSVT;} ;
  inline int getIPTTL() {return con->ipTTL;}; 
private:

  struct construct
  {
    int piLength ;
    int ipTTL ;
    int ipVer ;
    int reserved ;
    long RSVT;
    int peerIdentity ;
    nsaddr_t interfaceAddress;
  };
  struct construct *con;
};
// -------------------------------------------------- object Stack-Configuration-Data----------------------

class TlvObjectSCD: public TlvObject
{
public:
  virtual ~TlvObjectSCD(){};
  TlvObjectSCD(unsigned char *cont);
  TlvObjectSCD(int HLCount, int HLIprotocol, int HLIport);
  inline int getFirstHighLevelInterfaceProtocol() 	{return con->HLIprotocol;} ;
  inline int getFirstHighLevelInterfacePort() 		{return con->HLIport;} ;
  inline int getHLCount()				{return con->HLCount;};

private:

  struct construct
  { int HLCount;
    int HLIprotocol;
    int HLIport ;
  };
  struct construct *con;
};
// -------------------------------------------------- object ERROR -----------------------------------------

class TlvObjectERROR: public TlvObject
{
public:

  TlvObjectERROR(int errorClass, int errorCode, int esilength, char *errorSourceIdentifier,
                 int opterrorlength, char *optionalErrorSpecificInformation) ;

  inline void configErrorClass(int config) {con->errorClass = config;};
  inline void configErrorCode(int config) {con->errorCode =config;};
  inline int  get_ErrorClass() {return con->errorClass ;};
  inline int  get_ErrorCode() {return con->errorCode;};
  char  	*get_SourceIdentifier();
  char 	*get_optionalErrorSpecificInformation();
private:
  struct construct
  {
    int errorClass;
    int errorCode ;
    int esiLength ;
    int reserved  ;
    char *errorSourceIdentifier;
    char *optionalErrorSpecificInformation;
  };
  struct construct *con;
};

class TlvObjectNSLP : public TlvObject
{
public:

  TlvObjectNSLP(char *cont);
  
  virtual ~TlvObjectNSLP() { delete[] con->NSLP; };
  char *getNSLP();

private:
  struct construct
  {
    char *NSLP;
  };
  struct construct *con;
};

#endif
