/***************************************************************************
 *   Copyright (C) 2005 by Ribeiro, Romulo Eugenio                         *
 *   romulo@dei.uc.pt                                                      *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "nsis.h"
#include "rng.h"

static class NSLPClass : public TclClass
{
public:
  NSLPClass() : TclClass("Application/NSIS") {}
  TclObject* create(int, const char*const*)
  {  
    return (new NSLPApp());
    
  }
}
class_NSLP;

NSLPApp::NSLPApp(): NSLP_timer_(this)
{ 
  session_nslp = NULL;
  refresh_ = 25;
  last_data = NULL;
  bind("refresh_", &refresh_);
  bind("nslpid_", &nslpid_);
 
};

void NSLPApp::reset_NSLPApp_Timer()
{
  NSLP_timer_.resched((double)refresh_);
};

void NSLPApp_Timer::expire(Event *e)
{
  app_->refresh();
  //printf("\n\nNSLPApp TIMER EXPIRED\n\n");
  app_->reset_NSLPApp_Timer();
};

int NSLPApp::command(int argc, const char*const* argv)
{
  nsaddr_t n1, n2;
  int sii, res;
  Tcl& tcl = Tcl::instance();
  

  if (strcmp(argv[1], "attach-agent") == 0)
  { 
    agent_ = (Agent*) TclObject::lookup(argv[2]);
    if (agent_ == 0)
    {
      tcl.resultf("no such agent %s", argv[2]);
      return(TCL_ERROR);
    }
    agent_->attachApp(this);
    
    gimps_ = (GIMPSAgent*) TclObject::lookup(argv[2]);
    gimps_->nslp_ = this;
 
    return(TCL_OK);
  }
 
  /* session dest fid */
  if (strcmp(argv[1], "session") == 0)
  {    
    if (argc != 5)
    {
      tcl.result("Argmument number its not enough");
      return (TCL_ERROR);
    }
    if (argv[2][0] == '_')
    {
      tcl.evalf("%s id", argv[2]);					//here the ns2 provide a node object similar to _o13
      n1 = atoi(tcl.result());					//so we need to invoke the ns to return us the node's ID.
      tcl.evalf("%s id", argv[3]);
      n2 = atoi(tcl.result());
      sii= new_session_sid(n1, n2, atoi(argv[4]));
      tcl.resultf("%d ", sii);
    }
    else
    {
      tcl.resultf("%d", new_session_sid(strtol(argv[2], NULL, 0), strtol(argv[3], NULL, 0), atoi(argv[4])));
    }
    return (TCL_OK);
  };

  if (strcmp(argv[1], "sessions") == 0)
  { 
    this->list_sessions_agent();
    return (TCL_OK);
  };

  if (strcmp(argv[1], "start") == 0)
  {
    NSLP_timer_.resched(0.0);
    return (TCL_OK);
  };

   if (strcmp(argv[1], "sender") == 0)
  {
    if (argc != 4)
    {
      printf("The command <sender> dont have enought parameters \n Example: $nsis sender $ses1 000001 ");
      return (TCL_ERROR);
    }
    last_session = atoi(argv[2]);
    
    if(last_data != NULL) {
    	delete[] last_data;
    	last_data = NULL; 
    }
    last_data = new char[sizeof(argv[3])+1];
    strncpy(last_data, argv[3], sizeof(argv[3]));
    res = sendMessage(last_session, last_data); 
    if (res == TCL_OK)
    	NSLP_timer_.resched(refresh_);
    return res;
  };

  if (strcmp(argv[1], "refresh") == 0)
  {
    refresh();
    return (TCL_OK);
  };

  return (Application::command(argc, argv));

};

int NSLPApp::sendMessage(int session, char *data) {
    Tcl& tcl = Tcl::instance();
    NSLPmessage *msg = find_session_sid(session); //find the sessionid, build the MRI, SI and NSLP messages
    if (msg == NULL)
    {
      tcl.resultf("%s NSIS: unknown session ID %s", name(), session);
      return(TCL_ERROR);
    }
    printf("\n>>[NSLP Application] Node %d Sending NSLP Message. Session ID: %d  \n", 
    gimps_->gimpsIP, msg->si->getSessionId());
    gimps_->sendMessage(msg->mri, data, msg->si);
    delete msg;
    return (TCL_OK);
}

void NSLPApp::list_sessions_agent()
{
  session *ss = session_nslp;
  printf("\nSession set up in the node %d\n", gimps_->gimpsIP);
  while (ss!=NULL)
  {
    printf("\nSession %d    Source  %ld   Destination  %ld   Flow Id %d \n", 
    ss->sid, ss->src, ss->dst, ss->flowid);
    ss=ss->next;
  } ;

};

int NSLPApp::new_session_sid(nsaddr_t src, nsaddr_t dst, int fid)
{
  int sess = gimps_->getnewsession();
  session *ss = session_nslp;
  session *tail = session_nslp;

  while (ss!=NULL)
  {
    tail = ss;
    ss = ss->next;
  }
  ss = new session();
  ss->sid = sess;
  ss->dst = dst;
  ss->src = src;
  ss->flowid = fid;
  ss->next = NULL;
  if (tail == NULL)
  {
    session_nslp = ss;
  }
  else
  {
    tail->next=ss;
  }
  return sess;
}

NSLPmessage *NSLPApp::find_session_sid(int ssession )
{
  NSLPmessage *NSLPM;
  session *ss=session_nslp;
  session *tail=session_nslp;

  while  ( (ss!=NULL) && (ss->sid != ssession))
  {
    tail=ss;
    ss=ss->next;
  };
  if (ss==NULL) return NULL;

  TlvObjectMRI *MRI1 = new TlvObjectMRI(gimps_->messageRoutingMethod_,
  gimps_->gimpsIpVer,0,0,0,0,1,1,DOWNSTREAM,0,ss->src,ss->dst,0,0,DATAGRAM,
  0,0,0,0,1,1); 
  TlvObjectSI *SI = new TlvObjectSI (ss->sid, ss->flowid);
  NSLPM = new NSLPmessage(MRI1, SI, NULL, DOWNSTREAM, gimps_->NSLPID, gimps_->gimpsIP);
  return NSLPM;

};

void NSLPApp::recvnslp(NSLPmessage *nslpmesg)
{
    if (nslpmesg != NULL)
    {
      if (nslpmesg->directionflow == UPSTREAM)
      {
        if (nslpmesg->mri->getSourceAddress() != gimps_->gimpsIP)
        {
          gimps_->sendMessage(nslpmesg->mri, nslpmesg->NSLPData->getNSLP(), 
          nslpmesg->si);
          delete nslpmesg;
        }
      }
      else
      { 
   		printf("[NSLPApp recvnslp] Node %d >>> Nslp data message received: %s\n", 
        	gimps_->gimpsIP, nslpmesg->NSLPData->getNSLP());
        //executecmd(nslpmesg->NSLPData->getNSLP(), datagimps->fromhop);
        ///getchar();
        if (nslpmesg->mri->getDestinationAddress() != gimps_->gimpsIP)
        { 
          gimps_->sendMessage(nslpmesg->mri, nslpmesg->NSLPData->getNSLP(),
           nslpmesg->si);
           delete nslpmesg;
        }
        else {
        	printf("[NSLP Application] Node %d >>> Nslp data message arrived at destination %s\n", 
        	gimps_->gimpsIP);
        	delete nslpmesg;
        }
      }
    };

};

void NSLPApp::refresh()
{
  //double now = Scheduler::instance().clock();
  sendMessage(last_session, last_data);
  //  this->gimps_->checkAllMRS();
  //this->receivemsgnslp();
  //this->send_message_pendent();

};

void NSLPApp::executecmd(char *cmd, nsaddr_t from)
{
  if (strncmp(cmd, "set-nqueue", 9) == 0)
  {
    Tcl& tcl = Tcl::instance();
    tcl.evalf("[Simulator instance] %s %d %d", cmd, from ,gimps_->gimpsIP );
  }
}
